package nrs.encryption;

import java.io.*;

import org.bouncycastle.crypto.CryptoException;
import org.bouncycastle.crypto.digests.MD5Digest;
import org.bouncycastle.crypto.engines.BlowfishEngine;
import org.bouncycastle.crypto.modes.CBCBlockCipher;
import org.bouncycastle.crypto.params.KeyParameter;
import org.bouncycastle.crypto.paddings.PaddedBufferedBlockCipher;

/*
 * This file is part of JCards.
 * JCards is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * JCards is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JCards; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * Wrapper to encryption routines
 *
 * @author Copyright (C) 2002,2003 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class Encryptor extends OutputStream
{
    private PaddedBufferedBlockCipher cipher;
    private KeyParameter key;
	private OutputStream stream = null;
	private int totalSize=0, totalOlen=0, size=0, olen=0;
	private byte[] tmpresult;
	private boolean finished = false;
	private ByteArrayOutputStream result = null;

    /** Initialize the cryptographic engine */
    public Encryptor(byte[] password, boolean encrypt)
    {
        cipher = new PaddedBufferedBlockCipher(
                    new CBCBlockCipher(
                       new BlowfishEngine()));

        this.key = new KeyParameter(password);
		cipher.init(encrypt, key);
    }

    /** Initialize the cryptographic engine */
    public Encryptor(String key, boolean encrypt)
	{
        this(key.getBytes(), encrypt);
    }

	/**
	 * Initialize the cryptographic engine. Assumes decrypting.
	 * @see #Encryptor(String, boolean)
	 */
    public Encryptor(String key)
	{
		this(key.getBytes(), false);
	}

	public Encryptor(String key, OutputStream stream, boolean encrypt)
	{
		this(key.getBytes(), encrypt);
		setOutputStream(stream);
	}
	
	public void init()
	{
		totalSize = totalOlen = olen = size = 0;
		finished = false;
		result = new ByteArrayOutputStream();
	}

	public void setOutputStream(OutputStream stream)
	{
		this.stream = stream;
	}

    /**
	 * Routine that does the gritty work. Need to call doFinal at the end.
	 * @param data array of bytes to encode
	 * @see #doFinal
	 */
    public void callCipher(byte[] data) throws CryptoException, IOException
    {
        size = cipher.getOutputSize(data.length);
		totalSize += size;
        tmpresult = new byte[size];
        olen = cipher.processBytes(data, 0, data.length, tmpresult, 0);
		totalOlen += olen;
		
/*		if (olen < size)
		{
			byte[] tmp = new byte[olen];
			System.arraycopy(tmpresult, 0, tmp, 0, olen);
			tmpresult = tmp;
		}*/

		if (stream != null) {
			stream.write(tmpresult, 0, olen);
		} else {
			result.write(tmpresult, 0, olen);	
		}
    }

	/**
	 * Finishes encoding/decoding the data
	 */
	public void doFinal() throws CryptoException, IOException
	{
		if (finished) return;
		
		olen += cipher.doFinal(tmpresult, olen);
		totalOlen += olen;
		
		if (stream != null) {
			stream.write(tmpresult, 0, olen);
			stream.flush();
			stream.close();
		} else {
			if (olen < size)
			{
				byte[] tmp = new byte[olen];
				System.arraycopy(tmpresult, 0, tmp, 0, olen);
				tmpresult = tmp;
	        }
			result.write(tmpresult, 0, olen);
		}
		
		finished = true;
	}

    /** Encrypt arbitrary byte array, returning the
     * encrypted data in a different byte array */
    public synchronized byte[] encrypt(byte[] data) throws CryptoException, IOException
    {
        if (data == null || data.length == 0) {
			return new byte[0];
		}

		init();
        callCipher(data);
		doFinal();
		return result.toByteArray();
    }

    /** Decrypts arbitrary data */
    public synchronized byte[] decrypt(byte[] data) throws CryptoException, IOException
    {
        if (data == null || data.length == 0) {
            return new byte[0];
		}

		init();
        callCipher(data);
		doFinal();
		return result.toByteArray();
    }

	public byte[] getDigest()
	{
		MD5Digest digest = new MD5Digest();
		digest.update(key.getKey(), 0, key.getKey().length);
		byte[] out = new byte[digest.getDigestSize()];
		digest.doFinal(out, 0);

		return out;
	}

	public void write(byte[] b) throws IOException
	{
		try {
			callCipher(b);
		} catch (CryptoException cex) {
			throw new IOException(cex.getMessage());
		}
	}
	
	public void write(byte[] b, int off, int len) throws IOException
	{
		try {
			byte[] tmpbytes = new byte[len];
			System.arraycopy(b, off, tmpbytes, 0, len);
			callCipher(tmpbytes);
		} catch (CryptoException cex) {
			throw new IOException(cex.getMessage());
		}
	}
	
	public void write(int b) throws IOException
	{
		Integer i = new Integer(b);
		byte[] by = { i.byteValue() };
		try {
			callCipher(by);
		} catch (CryptoException cex) {
			throw new IOException(cex.getMessage());
		}
	}

	public void flush() throws IOException
	{
		try {
			doFinal();
		} catch (CryptoException cex) {
			throw new IOException(cex.getMessage());
		}
	}

	public void close() throws IOException
	{
		try {
			doFinal();
		} catch (CryptoException cex) {
			throw new IOException(cex.getMessage());
		}
		totalOlen = totalSize = olen = size = 0;
		result = null;
	}
}
