package nrs.jcards;

import java.awt.datatransfer.*;
import java.io.Serializable;
import java.util.*;

import nanoxml.*;
import nrs.jcards.io.Elements;

/*
 * This file is part of JCards.
 * JCards is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * JCards is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JCards; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * A card is one <i>page</i> of the database
 *
 * @author Copyright Copyright &#169; 2002-2004 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class Card implements Cloneable, Transferable
{
	private Hashtable values = new Hashtable();

	/** Create a new Card without any values */
	public Card()
	{}

	/**
	 * Create a new Card with the specified values
	 * @param newValues card values
	 */
	public Card(Hashtable newValues)
	{
		values = newValues;
	}

	/**
	 * Overwrite any existing values with the specified values
	 * @param newValues new card values
	 */
	public void setValues(Hashtable newValues)
	{
		values = newValues;
	}

	/**
	 * Add a value to this Card
	 * @param name field name
	 * @param val value of field
	 */
	public void addValue(String name, Object val)
	{
		if (val != null) {
			values.put(name, val);
		}
	}

	/**
	 * Get the value for the specified field
	 * @param fieldName name of the field
	 * @return the value as an Object or null if no value set
	 */
	public Object getValue(String fieldName)
	{
		return values.get(fieldName);
	}

	/**
	 * Get the combined values for the specified fields. Used by the Cards.quicksort method.
	 * @param firstField name of the first field
	 * @param secondField name of the second field which may be null
	 * @return the value as an Object or null if no value set
	 */
	public Object getValue(String firstField, String secondField)
	{
		String sval = null;
		Object val1 = getValue(firstField);
		if (val1 != null) sval = val1.toString();
		if (secondField != null)
		{
			Object val2 = getValue(secondField);
			if (val2 != null)
			{
				if (sval == null) sval = val2.toString();
				else sval += val2.toString();
			}
		}

		return sval;
	}

	/** Get the value as a String.
	 * Booleans are returned as Y/N.
	 * Memo fields are returned as the first line if truncateMemo is true.
	 * @param fieldName name of the field
	 * @param truncateMemo if true trucate memo fields
	 * @return the value as a String or null if no value set
	 */
	public String getStringValue(String fieldName, boolean truncateMemo)
	{
		Object obj = values.get(fieldName);

		if (obj == null)
		{
			return null;
		}
		else
		{
			DBDef def = DBDef.getInstance();
			FieldDef fd = def.getFieldDef(fieldName);
			switch (fd.getType())
			{
				case FieldDef.DATE:
					return def.formatDate((Date)obj);

				case FieldDef.BOOLEAN:
					Boolean b = (Boolean)obj;
					if (b.booleanValue()) return FieldDef.YES;
					else return FieldDef.NO;

				case FieldDef.MEMO:
					String memo = obj.toString();
					if (memo == null) return null;
					if (truncateMemo)
					{
						int nl = memo.indexOf("\n");
						if (nl > -1)
						{
							memo = memo.substring(0, nl);
						}
					}
					return memo;

				default:
					return obj.toString();
			}
		}
	}

	/**
	 * Get all values stored by this Card
	 * @return a Hashtable of values
	 */
	public Hashtable getValues()
	{
		return values;
	}

	/**
	 * Return the number of values stored by this Card
	 * @return number of values
	 */
	public int getNumberOfValues()
	{
		if (values == null) return 0;
		else return values.size();
	}

	/**
	 * Remove a value from the card
	 * @param fieldName field name to remove
	 */
	public void removeValue(String fieldName)
	{
		values.remove(fieldName);
	}

	/**
	 * Change a field name
	 * @param oldName the existing field name
	 * @param newName the new field name
	 */
	public void changeFieldName(String oldName, String newName)
	{
		Object val = values.remove(oldName);
		if (val != null) {
			values.put(newName, val);
		}
	}

	/**
	 * Change a field type. If the new and old field types are the same don't
	 * do anything. The new type can only be String or Memo so the value will
	 * be stored as a String.
	 * @param fieldName field name
	 * @param oldType old field type
	 * @param newType new field type
	 */
	public void changeFieldType(String fieldName, int oldType, int newType)
	{
		if (oldType == newType) return;

		String val = getStringValue(fieldName, false);
		if (oldType == FieldDef.MEMO) {
			val = val.replace('\n', ' ');
		}
		values.put(fieldName, val);
	}

	/**
	 * Get a String representation of this object suitable
	 * for saving to a file
	 * @return String version of this class
	 */
	public String toString()
	{
		StringBuffer sb = new StringBuffer();
		String fieldName;

		Enumeration enm = values.keys();
		while (enm.hasMoreElements())
		{
			fieldName = enm.nextElement().toString();
			sb.append(fieldName);
			sb.append("=");
			sb.append(values.get(fieldName));
			sb.append("\n");
		}

		return sb.toString();
	}

	/**
	 * Get the Card data as XML
	 * @return XML representation of the data
	 */
	public XMLElement getXML()
	{
		String field, fieldValue;
		XMLElement fieldEl;
		Object objValue;
		DBDef dbdef = DBDef.getInstance();

		Enumeration keys = values.keys();
		XMLElement cardEl = new XMLElement();
		cardEl.setTagName(Elements.CARD);
		while (keys.hasMoreElements())
		{
				field = keys.nextElement().toString();
				fieldEl = new XMLElement();
				fieldEl.setTagName(Elements.FIELD);
				fieldEl.addProperty(Elements.NAME, field);
				objValue = values.get(field);
				if (objValue instanceof Date) {
					fieldValue = dbdef.formatDate((Date)objValue);
				} else {
					fieldValue = objValue.toString().trim();
				}
				fieldEl.setContent(fieldValue);
				cardEl.addChild(fieldEl);
		}

		return cardEl;
	}

	/** Get Card data as text */
	public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException, java.io.IOException
	{
		return toString();
	}

	/** Get DataFlavour[] for clipboard transfer */
	public DataFlavor[] getTransferDataFlavors()
	{
		DataFlavor[] dfs = new DataFlavor[1];
		dfs[0] = DataFlavor.plainTextFlavor;
		return dfs;
	}

	/** Check to see if a DataFlavor is supported */
	public boolean isDataFlavorSupported(DataFlavor flavor)
	{
		if (flavor.equals(DataFlavor.plainTextFlavor)) return true;
		else return false;
	}

	/** Clone this object */
	protected Object clone() throws CloneNotSupportedException
	{
		return super.clone();
	}
}
