package nrs.jcards;

import java.io.*;
import java.text.*;
import java.util.*;

import jclass.util.JCVector;
import nanoxml.*;
import nrs.jcards.io.Elements;

/*
 * This file is part of JCards.
 * JCards is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * JCards is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JCards; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * Class to reference all the cards in a database.
 * This class is observed by the CardsList and observes
 * the DBDef.
 *
 * @author Copyright &#169; 2002-2004 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class Cards extends Observable implements Enumeration, Observer
{
	/** Action types */
	public final static int DELETE_CARD = 0;
	public final static int LOAD = 1;
	public final static int NEW_CARD = 2;
	public final static int UPDATE_CARD = 3;
	public final static int FIELD_ADD = 4;
	public final static int FIELD_CHANGE = 5;
	public final static int FIELD_DELETE = 6;
	public final static int SET_INFO = 7;
	public final static int FIND = 8;

	/** Sort directions */
	public final static int SORT_ASCENDING = 0;
	public final static int SORT_DESCENDING = 1;

	private String info = null;
	private Vector cards = new Vector();
	private Vector foundCards = null;
	private Card deletedCard = null;
	private JCards jcards;
	private int enumIndex = 0;

	public Cards(JCards jc)
	{
		jcards = jc;
	}

	/**
	 * Set some information about the cards.
	 * For example, "Music Database"
	 * @param newInfo the comment/note/info
	 */
	public void setInfo(String newInfo)
	{
		info = newInfo;
		setChanged();
		notifyObservers(new CardUpdate(SET_INFO));
	}

	/**
	 * Get any info that has been previously set
	 * @return info
	 */
	public String getInfo()
	{
		return info;
	}

	/**
	 * Used by data import to load without informing listeners
	 * @param cardXML the XMLElement that contains a Card
	 * @exception InvalidFieldException if the data is incorrect
	 */
	public void addCardSilently(XMLElement cardXML) throws InvalidFieldException
	{
		Card card = new Card();
		XMLElement fieldXML;
		String fieldName, value;
		Object objVal;
		FieldDef fd;
		DBDef dbdef = DBDef.getInstance();

		Vector children = cardXML.getChildren();
		Enumeration enm = children.elements();

		try
		{
			while (enm.hasMoreElements())
			{
				fieldXML = (XMLElement)enm.nextElement();
				if (!fieldXML.getTagName().equals(Elements.FIELD))
				{
					throw new InvalidFieldException("Card XML is incorrect");
				}

				fieldName = fieldXML.getProperty(Elements.NAME);
				fd = dbdef.getFieldDef(fieldName);
				value = fieldXML.getContents();
				objVal = FieldDef.parseValue(fd.getType(), value);
				if (objVal != null) {
					card.addValue(fieldName, objVal);
				}
			}
		} catch (ParseException pex) {
			throw new InvalidFieldException("Invalid field: "+pex.getMessage());
		} catch (NumberFormatException nfex) {
			throw new InvalidFieldException("Invalid field: "+nfex.getMessage());
		}

		cards.addElement(card);
	}

	/**
	 * Add a card without informing the listeners
	 * @param card the Card to add
	 */
	public void addCardSilently(Card card)
	{
		cards.addElement(card);
	}

	/**
	 * This method should be called to indicate that either a file or CSV
	 * read has finished and that the listeners can be informed.
	 * @see addCSVCard(String, char, char)
	 */
	public void finishedLoad()
	{
		sortCards();
		setChanged();
		notifyObservers(new CardUpdate(LOAD));
	}

	/**
	 * Add a card to the list.
	 * @param newCard the Card to add
	 * @return the index of the new card
	 */
	public int addCard(Card newCard)
	{
		JCards.log("Adding new card "+newCard, 9);

		cards.addElement(newCard);

		// add first field as visible
		DBDef def = DBDef.getInstance();
		Vector visFields = def.getVisibleFields();
		if (visFields.size() == 0)
		{
			String firstField = def.getFieldDef(0).getName();
			visFields.addElement(firstField);
		}
		sortCards();
		setChanged();
		int index = cards.indexOf(newCard);
		notifyObservers(new CardUpdate(NEW_CARD, index));

		return index;
	}

	/**
	 * Delete a card
	 * @param index index of the card
	 * @return the deleted card
	 */
	public void deleteCard(int index)
	{
		JCards.log("Deleting card at "+index, 9);

		Vector listCards = (foundCards == null ? cards : foundCards);
		deletedCard = (Card)listCards.elementAt(index);
		cards.removeElement(deletedCard);
		if (foundCards != null) foundCards.removeElement(deletedCard);
		setChanged();
		notifyObservers(new CardUpdate(DELETE_CARD, index));
	}

	/** Restore a previously deleted card */
	public void undeleteCard()
	{
		if (deletedCard != null)
		{
			JCards.log("Undeleting card", 9);

			cards.addElement(deletedCard);
			foundCards = null;
			sortCards();
			int index = cards.indexOf(deletedCard);
			deletedCard = null;
			setChanged();
			notifyObservers(new CardUpdate(NEW_CARD, index));
		}
	}

	/**
	 * Get the fields that will be displayed for all cards
	 * @return a JCVector of JCVector of field values
	 */
	public JCVector getRowEntries()
	{
		Card card;
		DBDef dbd = DBDef.getInstance();
		JCVector rows = new JCVector(cards.size()), row;
		Vector visFields = dbd.getVisibleFields();
		Enumeration enm, fenum;
		String fieldName;

		if (dbd.getNumberOfFields() > 0)
		{
			enm = (foundCards == null ? cards.elements() : foundCards.elements());
			while (enm.hasMoreElements())
			{
				card = (Card)enm.nextElement();
				row = new JCVector(visFields.size());
				fenum = visFields.elements();
				while (fenum.hasMoreElements())
				{
					fieldName = (String)fenum.nextElement();
					row.addElement(card.getStringValue(fieldName, true));
				}
				rows.addElement(row);
			}
		}

		return rows;
	}

	/**
	 * Get the fields that will be displayed for one card
	 * @return a JCVector of values
	 */
	public JCVector getRowEntries(int cardIndex)
	{
		if (cardIndex < 0) return null;

		Vector cardList = (foundCards == null ? cards : foundCards);
		Card card = (Card)cardList.elementAt(cardIndex);
		DBDef dbd = DBDef.getInstance();
		Vector visFields = dbd.getVisibleFields();
		String fieldName;
		JCVector row = new JCVector(visFields.size());
		Enumeration enm;

		if (dbd.getNumberOfFields() > 0)
		{
			enm = visFields.elements();
			while (enm.hasMoreElements())
			{
				fieldName = (String)enm.nextElement();
				row.addElement(card.getStringValue(fieldName, false));
			}
		}

		return row;
	}

	/**
	 * Get a Card
	 * @param index the index of the card to get
	 * @return the requested card or null if it doesn't exist
	 */
	public Card getCard(int index)
	{
		Vector cardList = (foundCards == null ? cards : foundCards);
		return (Card)cardList.elementAt(index);
	}

	/**
	 * Update a card definition
	 * @param index index of the card
	 * @param newCard new new Card definition
	 * @return the index of the updated card
	 */
	public int setCard(int index, Card newCard)
	{
		JCards.log("Updating card definition at "+index+" with "+newCard, 9);

		// just update the values referenced by the card in the right place
		Vector cardList = (foundCards == null ? cards : foundCards);
		Card c = (Card)cardList.elementAt(index);
		c.setValues(newCard.getValues());
		foundCards = null;
		sortCards();
		setChanged();
		notifyObservers(new CardUpdate(UPDATE_CARD, index));

		return cards.indexOf(c);
	}

	/**
	 * The underlying database definition has changed.
	 * Only deletes and changes need to be handled
	 */
	public void update(Observable o, Object arg)
	{
		DBDef def = DBDef.getInstance();

		if (arg != null)
		{
			String op = (String)arg;
			Card card;

			Enumeration enm = cards.elements();
			if (op.startsWith(DBDef.FIELD_DELETE))
			{
				JCards.log("Removing field from cards", 9);
				String fieldName = def.oldFieldDef.getName();
				// remove value for field from cards
				while (enm.hasMoreElements())
				{
					card = (Card)enm.nextElement();
					card.removeValue(fieldName);
				}

				if (op.equals(DBDef.FIELD_DELETE_VIS))
				{
					setChanged();
					notifyObservers(new CardUpdate(FIELD_DELETE));
				}
			}
			else if (op.equals(DBDef.FIELD_CHANGE))
			{
				JCards.log("Changing fields in cards", 9);
				String newFieldName = def.newFieldDef.getName();
				int newType = def.newFieldDef.getType();
				String oldFieldName = def.oldFieldDef.getName();
				int oldType = def.oldFieldDef.getType();

				while (enm.hasMoreElements())
				{
					card = (Card)enm.nextElement();
					card.changeFieldName(oldFieldName, newFieldName);
					card.changeFieldType(newFieldName, oldType, newType);
				}
				setChanged();
				notifyObservers(new CardUpdate(FIELD_CHANGE));
			}
			else if (op.equals(DBDef.FIELD_MOVE_FINISHED) || op.equals(DBDef.FIELD_SORT) || op.equals(DBDef.FIELD_ADD))
			{
				sortCards();
				setChanged();
				notifyObservers(new CardUpdate(LOAD));
			}
			else if (op.equals(DBDef.VISIBLE_FIELDS))
			{
				setChanged();
				notifyObservers(new CardUpdate(LOAD));
			}
		}
	}


	/**
	 * Get the data as XML
	 * @return the data as a XMLElement
	 */
	public XMLElement getXML()
	{
		Card c;
		Enumeration enm;
		XMLElement cardsEl = new XMLElement();
		cardsEl.setTagName(Elements.CARDS);

		if (info != null)
		{
			XMLElement infoTag = new XMLElement();
			infoTag.setTagName(Elements.INFO);
			infoTag.setContent(info);
			cardsEl.addChild(infoTag);
		}

		enm = cards.elements();
		while (enm.hasMoreElements())
		{
			c = (Card)enm.nextElement();
			cardsEl.addChild(c.getXML());
		}

		return cardsEl;
	}

	/**
	 * Read previously saved data
	 * @param cardsXML the import data
	 * @exception InvalidFieldException field data is incorrect
	 */
	public void setXML(XMLElement cardsXML) throws InvalidFieldException
	{
		String tagName;
		XMLElement child;

		// remove existing data
		info = null;
		cards = new Vector();
		foundCards = null;

		Vector children = cardsXML.getChildren();
		Enumeration enm = children.elements();
		while (enm.hasMoreElements())
		{
			child = (XMLElement)enm.nextElement();
			tagName = child.getTagName();

			if (tagName.equalsIgnoreCase(Elements.INFO))
			{
				info = child.getContents();
			}
			else if (tagName.equalsIgnoreCase(Elements.CARD))
			{
				addCardSilently(child);
			}
		}

		// see if the first field is visible
		if (cards.size() > 0)
		{
			DBDef def = DBDef.getInstance();
			Vector visFields = def.getVisibleFields();
			if (visFields.size() == 0)
			{
				String firstField = def.getFieldDef(0).getName();
				visFields.addElement(firstField);
			}
		}

		sortCards();
		setChanged();
		notifyObservers(new CardUpdate(LOAD));
	}

	/**
	 * Get the data in CSV format
	 * @param fieldDelim field delimiter
	 * @param textDelim text delimiter
	 * @param concatNL if true new lines are removed from the text
	 * @return CSV data
	 */
	public String getCSV(String fieldDelim, String textDelim, boolean concatNL)
	{
		StringBuffer sb = new StringBuffer();
		Card c;
		Enumeration cardEnum, fieldEnum;
		Object val;
		String field, sval;
		Date dval;
		DBDef dbdef = DBDef.getInstance();

		cardEnum = cards.elements();
		while (cardEnum.hasMoreElements())
		{
			c = (Card)cardEnum.nextElement();
			fieldEnum = dbdef.getFieldNames().elements();
			while (fieldEnum.hasMoreElements())
			{
				field = (String)fieldEnum.nextElement();
				val = c.getValue(field);
				if (val == null)
				{
					sb.append(textDelim);
					sb.append(textDelim);
				}
				else
				{
					// remove newlines
					if (val instanceof String)
					{
						sval = textDelim+(String)val+textDelim;
						if (concatNL) {
							sval = sval.replace('\n', ' ');
						}
					}
					else if (val instanceof Date)
					{
						dval = (Date)val;
						sval = dbdef.formatDate(dval);
					}
					else
					{
						sval = val.toString();
					}

					sb.append(sval);
				}
				sb.append(fieldDelim);
			}
			// remove last comma
			sb.setLength(sb.length()-1);
			sb.append("\n");
		}
		// remove last new line
		sb.setLength(sb.length()-1);

		return sb.toString();
	}

	/** Sort on a particular field */
	public void sortOnField(String fieldName, int direction)
	{
		DBDef.getInstance().setSortCriteria(fieldName, null, direction);
		sortCards();
		setChanged();
		notifyObservers(new CardUpdate(LOAD));
	}

	/** Sort the Cards into ascending order */
	private void sortCards()
	{
		int i;

		if (cards.size() > 0)
		{
			Card cardArray[] = new Card[cards.size()];
			// convert to an array
			for (i=0; i<cardArray.length; i++)
			{
				cardArray[i] = (Card)cards.elementAt(i);
			}
			String sortField = DBDef.getInstance().getSortField();
			String secondSortField = DBDef.getInstance().getSecondSortField();
			try
			{
				quicksort(cardArray, 0, cardArray.length-1, sortField, secondSortField);

				cards = new Vector(cardArray.length);
				if (DBDef.getInstance().getSortDirection() == SORT_ASCENDING)
				{
					for (i=0; i<cardArray.length; i++)
					{
						cards.addElement(cardArray[i]);
					}
				}
				else
				{
					for (i=cardArray.length; i>0; i--)
					{
						cards.addElement(cardArray[i-1]);
					}
				}
			} catch (NullPointerException npex)	{
				// it's possible that a new mandatory field
				// has been added but the default values
				// are still null
				System.err.println("Error sorting - null field");
			}
		}
	}

	/**
	 * Sort the Cards into ascending order
	 * @param cardArray the array of Cards to sort
	 * @param lo0 start
	 * @param hi0 end
	 * @param sortField field to sort on
	 * @param secondSortField second field to sort on - may be null
	 */
	private void quicksort(Card[] cardArray, int lo0, int hi0, String sortField, String secondSortField)
	{
		Card c;
		int lo = lo0;
		int hi = hi0;
		int tmp;
		Object val = null;
		String mid = "";

		if (lo >= hi) {
			return;
		}

		// get the mid point
		val = cardArray[(lo + hi) / 2].getValue(sortField, secondSortField);
		mid = (val == null ? "" : val.toString());

		while (lo < hi)
		{
			while (lo<hi)
			{
				val = cardArray[lo].getValue(sortField, secondSortField);
				if (val == null) val = "";
				if (val.toString().compareTo(mid) >= 0) break;
				lo++;
			}
			while (lo<hi)
			{
				val = cardArray[hi].getValue(sortField, secondSortField);
				if (val == null) val = "";
				if (val.toString().compareTo(mid) <= 0) break;
				hi--;
			}
			if (lo < hi)
			{
				c = cardArray[lo];
				cardArray[lo] = cardArray[hi];
				cardArray[hi] = c;
				lo++;
				hi--;
			}
		}
		if (hi < lo)
		{
			tmp = hi;
			hi = lo;
			lo = tmp;
		}
		quicksort(cardArray, lo0, lo, sortField, secondSortField);
		quicksort(cardArray, lo==lo0?lo+1:lo, hi0, sortField, secondSortField);
	}

	/**
	 * Search for matching cards
	 * @param search the search text
	 * @param either 'All' or a particular field
	 * @param matchCase set to true for a case sensitive search
	 * @param searchInResults if true use the existing search results as the starting point
	 * @return true if any cards are found, false otherwise
	 */
	public boolean findCards(String search, String field, boolean matchCase, boolean searchInResults)
	{
		Card c;
		Hashtable fhash;
		Object val;
		String fieldValue;
		String lowerSearch = search.toLowerCase();
		Enumeration cenum, fenum=null;
		boolean matchAllFields = false;
		Vector findCards = new Vector();
		ResourceBundle menu = JCards.getMenuBundle();
		String all = menu.getString("all");
		menu = null;

		if (field.equals(all)) matchAllFields = true;

		if (searchInResults && foundCards != null && foundCards.size() > 0) {
			cenum = foundCards.elements();
		} else {
			cenum = cards.elements();
		}

		while (cenum.hasMoreElements())
		{
			c = (Card)cenum.nextElement();

			if (matchAllFields)
			{
				fhash = c.getValues();

				fenum = fhash.elements();
				while (fenum.hasMoreElements())
				{
					fieldValue = fenum.nextElement().toString();
					if (matchCase && fieldValue.indexOf(search) > -1)
					{
						if (!findCards.contains(c)) {
							findCards.addElement(c);
						}
					}
					else if (!matchCase && fieldValue.toLowerCase().indexOf(lowerSearch) > -1)
					{
						if (!findCards.contains(c)) {
							findCards.addElement(c);
						}
					}
				}
			}
			else
			{
				val = c.getValue(field);
				if (val != null)
				{
					fieldValue = val.toString();
					if (matchCase && fieldValue.indexOf(search) > -1)
					{
						findCards.addElement(c);
					}
					else if (!matchCase && fieldValue.toLowerCase().indexOf(lowerSearch) > -1)
					{
						findCards.addElement(c);
					}
				}
			}
		}

		if (findCards.size() > 0)
		{
			foundCards = findCards;
			setChanged();
			notifyObservers(new CardUpdate(FIND));
		}
		else
		{
			foundCards = null;
		}

		return (findCards.size() > 0);
	}

	/**
	 * Reset the search results
	 */
	public void resetFind()
	{
		if (foundCards != null)
		{
			foundCards = null;
			setChanged();
			notifyObservers(new CardUpdate(FIND));
		}
	}

	/**
	 * Return the number of cards
	 * @return number of cards
	 */
	public int size()
	{
		return cards.size();
	}

	/** Clear existing data */
	public void reset()
	{
		cards.removeAllElements();
		foundCards = null;
		deletedCard = null;
		info = null;
	}

	/**
	 * Get date statistics for the specified field
	 * @param fieldName field to collect statistics on
	 * @return statistics as an array of Strings
	 */
	public String[] getDateStatistics(String fieldName)
	{
		Card card;
		Date date, earliest=new Date(), latest=new Date(), average;

		date = (Date)getCard(0).getValue(fieldName);
		earliest = date;
		latest = date;
		average = date;

		for (int i=0; i<cards.size(); i++)
		{
			card = getCard(i);
			date = (Date)card.getValue(fieldName);

			if (date != null)
			{
				if (earliest == null) {
					earliest = date;
					latest = date;
				}
				if (date.before(earliest)) earliest = date;
				else if (date.after(latest)) latest = date;
			}
		}

		// check for no data
		if (earliest == null) return null;

		// calculate median date
		long diff = latest.getTime() - earliest.getTime();
		if (diff > 0)
		{
			diff /= 2;
			average = new Date(earliest.getTime() + diff);
		}
		else {
			average = earliest;
		}

		DBDef def = DBDef.getInstance();
		MessageFormat eFormat = new MessageFormat(jcards.getMessage("stats.earliest"));
		MessageFormat lFormat = new MessageFormat(jcards.getMessage("stats.latest"));
		MessageFormat mFormat = new MessageFormat(jcards.getMessage("stats.median"));
		String[] stats = new String[3];
		String[] args = { def.formatDate(earliest) };
		stats[0] = eFormat.format(args);
		args[0] = def.formatDate(latest);
		stats[1] = lFormat.format(args);
		args[0] = def.formatDate(average);
		stats[2] = mFormat.format(args);

		return stats;
	}

	/**
	 * Get numeric statistics for the specified field
	 * @param fieldName field to collect statistics on
	 * @return statistics as an array of Strings
	 */
	public String[] getNumericStatistics(String fieldName)
	{
		Card card;
		Number num;
		double val, total = 0, average;
		double min=Double.MAX_VALUE, max=Double.MIN_VALUE;
		int dataCount = 0;

		num = (Number)getCard(0).getValue(fieldName);
		if (num != null) {
			min = max = num.doubleValue();
		}

		for (int i=0; i<cards.size(); i++)
		{
			card = getCard(i);
			num = (Number)card.getValue(fieldName);

			if (num != null)
			{
				val = num.doubleValue();
				if (val < min) min = val;
				if (val > max) max = val;

				total += val;
				dataCount++;
			}
		}

		// check for no data
		if (dataCount == 0) return null;

		average = total / dataCount;

		String[] stats = new String[4];
		MessageFormat msgFormat = new MessageFormat(jcards.getMessage("stats.min"));
		Double[] args = { new Double(min) };
		stats[0] = msgFormat.format(args);
		msgFormat.applyPattern(jcards.getMessage("stats.max"));
		args[0] = new Double(max);
		stats[1] = msgFormat.format(args);
		msgFormat.applyPattern(jcards.getMessage("stats.total"));
		args[0] = new Double(total);
		stats[2] = msgFormat.format(args);
		msgFormat.applyPattern(jcards.getMessage("stats.avg"));
		args[0] = new Double(average);
		stats[3] = msgFormat.format(args);

		return stats;
	}

	/** Return an Enumeration over the Card objects */
	public Enumeration elements()
	{
		enumIndex = 0;
		return this;
	}

	/** Return true if there are more elements to enumerate over */
	public boolean hasMoreElements()
	{
		return enumIndex < cards.size();
	}

	/**
	 * Get the next element in the Enumeration
	 * @return next Object
	 */
	public Object nextElement()
	{
		return cards.elementAt(enumIndex++);
	}
}
