package nrs.jcards;

import java.awt.event.ActionEvent;
import java.util.Observable;
import java.util.Observer;
import java.util.Vector;
import jclass.bwt.JCActionEvent;
import jclass.bwt.JCActionListener;
import jclass.bwt.JCMultiColumnList;
import jclass.util.JCVector;
import nanoxml.XMLElement;
import nrs.jcards.io.Elements;

/*
 * JCards
 * Copyright (C) 2004 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * A list that is aware of changes to the underlying data
 * structure.
 *
 * @author Copyright &#169; 2002-2004 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class CardsList extends JCMultiColumnList implements Observer, JCActionListener
{
	private JCards jcards = null;
	private String sortCol = null;
	private int sortDir = Cards.SORT_ASCENDING;
	
	public CardsList(JCards jcards)
	{
		super();
		
		this.jcards = jcards;
		
		setFont(jcards.baseFont);
		
		addActionListener(this);
		getHeader().addActionListener(this);
		setAutoSelect(true);
	}
	
	/**
	 * The underlying data structure has changed
	 */
	public void update(Observable o, Object arg)
	{
		if (arg != null)
		{
			CardUpdate cu = (CardUpdate)arg;
			switch (cu.cardOperation)
			{
				case Cards.LOAD:
				case Cards.FIELD_DELETE:
				case Cards.UPDATE_CARD:
				case Cards.FIND:
					displayCards();
					break;
					
				case Cards.NEW_CARD:
					addItem(jcards.getCards().getRowEntries(cu.cardIndex), cu.cardIndex);
					break;
					
				case Cards.DELETE_CARD:
					deleteItem(cu.cardIndex);
					break;
					
				case Cards.FIELD_CHANGE:
					String[] vf = DBDef.getInstance().getVisibleFieldsArray();
					setColumnButtons(vf);
					break;
					
				case Cards.SET_INFO:
					jcards.setChanges(true);
					break;
			}
		}
	}
	
	/** Display the cards in the list */
	private void displayCards()
	{
		int c = 0;
		int selection = getSelectedIndex();
		String cardName;
		Cards cards = jcards.getCards();
		
		setBatched(true);
		JCVector rows = cards.getRowEntries();
		if (rows.size() > 0) {
			setItems(rows);
		} else {
			deleteItems(0, countItems());
		}
		String[] cols = DBDef.getInstance().getVisibleFieldsArray();
		setColumnButtons(cols);
		setNumColumns(cols.length);
		
		// reselect any previously selected card
		if (selection>-1 && selection <= cards.size())
		{
			select(selection);
		}
		setBatched(false);
	}
	
	/**
	 * A Card was selected or a column was clicked
	 */
	public void actionPerformed(JCActionEvent e)
	{
		if (e.getSource() instanceof JCMultiColumnList)
		{
			// open a card
			jcards.editCard(false, jcards.isAutoEdit(), false);
		}
		else
		{
			if ((e.getModifiers() & ActionEvent.CTRL_MASK) == ActionEvent.CTRL_MASK) {
				// change visible fields
				jcards.editPrefs(true);
			}
			else
			{
				// set sort direction/column
				String col = e.getActionCommand();
				if (sortCol == null || !sortCol.equals(col))
				{
					sortCol = col;
					sortDir = Cards.SORT_ASCENDING;
				}
				else
				{
					if (sortDir == Cards.SORT_ASCENDING) {
						sortDir = Cards.SORT_DESCENDING;
					} else {
						sortDir = Cards.SORT_ASCENDING;
					}
				}

				jcards.getCards().sortOnField(sortCol, sortDir);
			}
		}
	}
	
	/**
	 * Get the column widths as an XMLElement
	 * @return XML representation of the column widths
	 */
	public XMLElement getColWidths()
	{
		XMLElement colEl = new XMLElement();
		colEl.setTagName(Elements.COL_WIDTHS);
		
		int[] colWidths = getColumnWidths();
		XMLElement width;
		for (int i=0; i<colWidths.length; i++)
		{
			width = new XMLElement();
			width.setTagName(Elements.WIDTH);
			width.setContent(Integer.toString(colWidths[i]));
			colEl.addChild(width);
		}
		
		return colEl;
	}
	
	/** Set the column widths */
	public void setColWidths(XMLElement colWidths)
	{
		XMLElement width;
		
		Vector children = colWidths.getChildren();
		int[] cw = new int[children.size()];
		try
		{
			for (int i=0; i<children.size(); i++)
			{
				width = (XMLElement)children.elementAt(i);
				cw[i] = Integer.parseInt(width.getContents());
			}

			setColumnWidths(cw);
			setVisibleRows(cw.length);
		} catch (Exception ex) {
			// NumberFormatException or NullPointerException
		}
	}
	
	/** Reset the list */
	public void reset()
	{
		clear();
		setColumnButtons(new String[0]);
	}	
}
