package nrs.jcards;

import java.text.*;
import java.util.*;

import nanoxml.XMLElement;

import nrs.jcards.io.DBDefXML;
import nrs.util.PostfixCalculator;

/*
 * This file is part of JCards.
 * JCards is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * JCards is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JCards; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * Contains the definition of the fields in the database and preferences.
 * Observers of this object are informed about changes: Cards.
 *
 * @author Copyright &#169; 2002-2004 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class DBDef extends Observable
{
	private static DBDef instance = null;
	private Vector fieldDefs = new Vector();
	private Vector visibleFields = new Vector();
	private Hashtable categories = new Hashtable();
	private Hashtable counters = new Hashtable();
	private String dateFormat = UserSettings.DATE_FORMATS[0];
	private SimpleDateFormat sdf = new SimpleDateFormat(dateFormat);
	private boolean prepopulateDate = false;
	private boolean displayFieldNames = true;
	private String sortField = null;
	private String secondSortField = null;
	private int sortDirection = Cards.SORT_ASCENDING;
	private static PostfixCalculator calculator;
	FieldDef oldFieldDef, newFieldDef;

	/** Action types */
	public final static String CATEGORY_ITEM_ADDED	= "CategoryItemAdded";
	public final static String CATEGORY_ITEM_MOVED	= "CategoryItemMoved";
	public final static String CATEGORY_ITEM_REMOVED	= "CategoryItemRemoved";
	public final static String CATEGORY_ITEM_RENAMED = "CategoryItemRenamed";
	public final static String CATEGORY_REMOVED		= "CategoryRemoved";
	public final static String CATEGORY_RENAMED		= "CategoryRenamed";
	public final static String CATEGORY_UPDATED		= "CategoryUpdated";
	public final static String COUNTER_ADDED = "CounterAdded";
	public final static String COUNTER_DELETED = "CounterDeleted";
	public final static String COUNTER_UPDATED = "CounterUpdated";
	public final static String FIELD_ADD		= "FieldAdd";
	public final static String FIELD_CHANGE		= "FieldChange";
	public final static String FIELD_DELETE		= "FieldDelete";
	public final static String FIELD_DELETE_VIS		= "FieldDeleteVisible";
	public final static String FIELD_MOVE		= "FieldMove";
	public final static String FIELD_MOVE_FINISHED	= "FieldMoveFinished";
	public final static String FIELD_SORT		= "FieldSort";
	public final static String RELOAD			= "Reload";
	public final static String VISIBLE_FIELDS		= "VisibleFields";

	private DBDef()
	{}

	/**
	 * Singleton patten for getting only one instance of DBDef.
	 * It isn't synchronized as there is no threading in JCards.
	 */
	public static DBDef getInstance()
	{
		if (instance == null) {
			instance = new DBDef();
		}

		return instance;
	}

	/**
	 * Singleton patten for getting only one instance of a PostfixCalculator.
	 * It isn't synchronized as there is no threading in JCards.
	 */
	public static PostfixCalculator getCalculator()
	{
		if (calculator == null) {
			calculator = new PostfixCalculator();
		}

		return calculator;
	}

	/**
	 * Add a field definition to the database. Field names must be unique.
	 * @param newFieldDef the field definition to add
	 */
	public boolean addFieldDef(FieldDef newFD)
	{
		JCards.log("Adding new field "+newFD, 7);

		// check to see if the field name already exists
		String name = newFD.getName();

		FieldDef fd;
		Enumeration enm = fieldDefs.elements();
		while (enm.hasMoreElements())
		{
			fd = (FieldDef)enm.nextElement();
			if (fd.getName().equals(name)) {
				JCards.log("Field definition already exists", 3);
				return false;
			}
		}

		// first field is always mandatory
		if (fieldDefs.size() == 0) {
			newFD.setMandatory(true);
			JCards.log("First field so making mandatory", 5);
		}

		fieldDefs.addElement(newFD);
		newFieldDef = newFD;

		// update the visible fields if empty
		if (visibleFields.size() == 0)
		{
			visibleFields.addElement(newFD.getName());
			sortField = newFD.getName();
		}

		setChanged();
		notifyObservers(FIELD_ADD);

		return true;
	}

	/**
	 * Remove a field definition
	 * @param index the index of the field that has been deleted
	 */
	public void removeFieldDef(int index)
	{
		JCards.log("Removing field "+index, 5);
		oldFieldDef = (FieldDef)fieldDefs.elementAt(index);

		fieldDefs.removeElementAt(index);
		boolean visible = visibleFields.removeElement(oldFieldDef.getName());
		// make sure that next the field - which is now the first - is mandatory
		FieldDef nextField = (FieldDef)fieldDefs.elementAt(0);
		if (nextField != null) {
			nextField.setMandatory(true);
			if (visibleFields.size() == 0) {
				visibleFields.addElement(nextField.getName());
			}
		}

		setChanged();
		if (visible) {
			notifyObservers(FIELD_DELETE_VIS);
		} else {
			notifyObservers(FIELD_DELETE);
		}
	}

	/**
	 * Get the number of fields
	 */
	public int getNumberOfFields()
	{
		return fieldDefs.size();
	}

	/**
	 * Get a FieldDef based on its position
	 * @param fieldNum field position
	 * @return the FieldDef
	 */
	public FieldDef getFieldDef(int fieldNum)
	{
		return (FieldDef)fieldDefs.elementAt(fieldNum);
	}

	/**
	 * Get a FieldDef based on its name
	 * @param fieldName field name
	 * @return the FieldDef or null if it doesn't exist
	 */
	public FieldDef getFieldDef(String fieldName)
	{
		FieldDef fd = null;

		if (fieldName == null) return null;

		Enumeration enm = fieldDefs.elements();
		while (enm.hasMoreElements())
		{
			fd = (FieldDef)enm.nextElement();
			if (fieldName.equals(fd.getName())) return fd;
		}

		return null;
	}

	/**
	 * Get a reference to the FieldDefs
	 * @return Vector of FieldDef objects
	 */
	public Vector getFieldDefs()
	{
		return fieldDefs;
	}

	/**
	 * Set the field definitions
	 * @param fieldDefs Vector of FieldDef objects
	 */
	public void setFieldDefs(Vector fieldDefs)
	{
		this.fieldDefs = fieldDefs;
	}

	/**
	 * Move a fields position in the list.
	 * @param index current position of the field
	 * @param incr whether to move the field up or down
	 * @param the new position of the field
	 */
	public int moveField(int index, int incr)
	{
		int pos = 0;
		FieldDef fd;

		JCards.log("Moving field "+index, 5);

		// start or end of list
		if (index == 0 && incr == -1) return 0;
		else if (incr == 1 && ((index+1) == fieldDefs.size())) return fieldDefs.size();

		fd = (FieldDef)fieldDefs.elementAt(index);
		fieldDefs.removeElementAt(index);

		if ((index+incr) > fieldDefs.size())
		{
			fieldDefs.addElement(fd);
			pos = fieldDefs.size()-1;
		}
		else
		{
			fieldDefs.insertElementAt(fd, index+incr);
			// first field is always mandatory
			if ((index+incr) == 0) fd.setMandatory(true);
			pos = index + incr;
		}

		setChanged();
		notifyObservers(FIELD_MOVE);

		return pos;
	}

	/**
	 * Convenience method to indicate that all moves have finished for now.
	 * All observers are notified.
	 */
	public void moveFieldFinished()
	{
		setChanged();
		notifyObservers(FIELD_MOVE_FINISHED);
	}

	/**
	 * Replace a field definition
	 * @param index position of the field
	 * @param newFieldDef the new field definition
	 */
	public void replaceFieldDef(int index, FieldDef newFD)
	{
		JCards.log("Replacing field "+index, 5);

		oldFieldDef = (FieldDef)fieldDefs.elementAt(index);
		newFieldDef = newFD;

		fieldDefs.removeElementAt(index);
		fieldDefs.insertElementAt(newFD, index);

		String fieldName = oldFieldDef.getName();
		int visIndex = visibleFields.indexOf(fieldName);
		if (visIndex > -1)
		{
			visibleFields.removeElementAt(visIndex);
			visibleFields.insertElementAt(newFieldDef.getName(), visIndex);
		}

		setChanged();
		notifyObservers(FIELD_CHANGE);
	}

	/**
	 * Add/modify a category
	 * @param catName category name
	 * @param catItems Vector of items
	 */
	public void setCategory(String catName, Vector catItems)
	{
		categories.put(catName, catItems);
		setChanged();
		notifyObservers(CATEGORY_UPDATED);
	}

	/**
	 * Get a Category or null if it doesn't exist
	 * @param catName Category name
	 */
	public Vector getCategory(String catName)
	{
		return (Vector)categories.get(catName);
	}

	/**
	 * Return a list of all categories
	 * @return the Hashtable of categories
	 */
	public Hashtable getCategories()
	{
		return categories;
	}

	/**
	 * Set the categories
	 * @param cats Hashtable of categories
	 */
	public void setCategories(Hashtable cats)
	{
		categories = cats;
		setChanged();
		notifyObservers(CATEGORY_UPDATED);
	}

	/**
	 * Return a list of counters
	 * @return the Hashtable of counters
	 */
	public Hashtable getCounters()
	{
		return counters;
	}

	/**
	 * Add a counter
	 * @param counter new Counter
	 */
	public void addCounter(Counter counter)
	{
		counters.put(counter.getName(), counter);
		setChanged();
		notifyObservers(COUNTER_ADDED);
	}

	/**
	 * Set the counters
	 * @param counts Hashtable of counters
	 */
	public void setCounters(Hashtable counts)
	{
		counters = counts;
		setChanged();
		notifyObservers(COUNTER_ADDED);
	}

	/**
	 * Delete a counter if it is not in use
	 * @param name name of the counter
	 * @return either return null if the counter isn't in use or the name of the field using this counter
	 */
	public String removeCounter(String name)
	{
		FieldDef fd;

		// check if in use by field
		Enumeration enm = fieldDefs.elements();
		while (enm.hasMoreElements())
		{
			fd = (FieldDef)enm.nextElement();
			if (fd.getType() == FieldDef.COUNTER && fd.getCounter().equals(name)) {
				return fd.getName();
			}
		}

		counters.remove(name);
		setChanged();
		notifyObservers(COUNTER_DELETED);

		return null;
	}

	/** Update a counter */
	public void updateCounter(Counter counter)
	{
		counters.put(counter.getName(), counter);

		setChanged();
		notifyObservers(COUNTER_UPDATED);
	}

	/** Get a Counter */
	public Counter getCounter(String counterName)
	{
		return (Counter)counters.get(counterName);
	}

	/**
	 * Are any counters defined
	 * @return true if the database has counters
	 */
	public boolean areCountersDefined()
	{
		if (counters.size() > 0) return true;
		else return false;
	}

	/**
	 * Are any calculated fields defined
	 * @return true if the database has calculations
	 */
	public boolean areCalculationsDefined()
	{
		FieldDef fd;
		Enumeration enm = fieldDefs.elements();
		while (enm.hasMoreElements())
		{
			fd = (FieldDef)enm.nextElement();
			if (fd.getType() == FieldDef.CALCULATED) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Calculated fields are only allowed if there are numeric fields already defined
	 * @return true if calculated fields are allowed
	 */
	public boolean areCalculationsAllowed()
	{
		FieldDef fd;
		Enumeration enm = fieldDefs.elements();
		while (enm.hasMoreElements())
		{
			fd = (FieldDef)enm.nextElement();
			if (fd.getType() == FieldDef.INTEGER || fd.getType() == FieldDef.DOUBLE || fd.getType() == FieldDef.COUNTER) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Returns data for CSV file
	 * @param fieldDelimiter the delimiter used to separate field names
	 * @return CSV data
	 */
	public String getCSV(String fieldDelimiter)
	{
		StringBuffer sb = new StringBuffer();
		FieldDef fd;

		Enumeration enm = fieldDefs.elements();
		while (enm.hasMoreElements())
		{
			fd = (FieldDef)enm.nextElement();
			sb.append(fd.getName());
			sb.append(fieldDelimiter);
		}

		// remove last comma
		if (sb.length() > 0) sb.setLength(sb.length()-1);

		return sb.toString();
	}

	/**
	 * Get a list of all the field names
	 * @return a Vector of field names
	 * @see getFieldNamesArray()
	 */
	public Vector getFieldNames()
	{
		Vector fieldNames = new Vector(fieldDefs.size());
		FieldDef fd;

		Enumeration enm = fieldDefs.elements();
		while (enm.hasMoreElements())
		{
			fd = (FieldDef)enm.nextElement();
			fieldNames.addElement(fd.getName());
		}

		return fieldNames;
	}

	/**
	 * Get field names as an array
	 * @see #getFieldNames()
	 */
	public String[] getFieldNamesArray()
	{
		Vector fieldNames = getFieldNames();
		String[] fnArray = new String[fieldNames.size()];
		int index = 0;
		Enumeration enm = fieldNames.elements();
		while (enm.hasMoreElements())
		{
			fnArray[index++] = (String)enm.nextElement();
		}
		return fnArray;
	}

	/** Delete all field definitions and reset the date format */
	public void reset()
	{
		fieldDefs.removeAllElements();
		categories.clear();
		counters.clear();
		visibleFields.removeAllElements();
		oldFieldDef = null;
		newFieldDef = null;
		dateFormat = UserSettings.DATE_FORMATS[0];
		prepopulateDate = false;
		sortField = null;
		secondSortField = null;
		displayFieldNames = true;
	}

	/**
	 * Remove a category
	 * @param catNam category name
	 * @return true if the category was removed, false if a field depends on it
	 */
	public boolean removeCategory(String catName)
	{
		if (catName != null)
		{
			JCards.log("Removing category "+catName, 5);

			// check to see if any fields use this category
			FieldDef fd;
			Enumeration enm = fieldDefs.elements();
			while (enm.hasMoreElements())
			{
				fd = (FieldDef)enm.nextElement();
				if (fd.getType() == FieldDef.CATEGORY && fd.getCategory().equals(catName))
				{
					return false;
				}
			}

			// they don't so remove it
			categories.remove(catName);
			setChanged();
			notifyObservers(CATEGORY_REMOVED);

			return true;
		}

		return false;
	}

	public void renameCategory(String oldCatName, String newCatName)
	{
		if (oldCatName != null && newCatName != null)
		{
			JCards.log("Renaming category "+oldCatName+" to "+newCatName, 5);

			Vector catItems = (Vector)categories.get(oldCatName);
			categories.put(newCatName, catItems);
			categories.remove(oldCatName);

			// update any FieldDefs
			FieldDef fd;
			Enumeration enm = fieldDefs.elements();
			while (enm.hasMoreElements())
			{
				fd = (FieldDef)enm.nextElement();
				if (fd.getType() == FieldDef.CATEGORY && fd.getCategory().equals(oldCatName))
				{
					fd.setCategory(newCatName);
				}
			}
			setChanged();
			notifyObservers(CATEGORY_RENAMED);
		}
	}

	/**
	 * Add a new category item
	 * @param catName category name
	 * @param item category item
	 */
	public void addCategoryItem(String catName, String item)
	{
		JCards.log("Adding "+item+" to category "+catName, 5);
		Vector catItems = (Vector)categories.get(catName);
		if (catItems != null)
		{
			catItems.addElement(item);
			setChanged();
			notifyObservers(CATEGORY_ITEM_ADDED);
		}
	}

	/**
	 * Rename a category item
	 * @param catName category name
	 * @param oldItem old item name
	 * @param newItem new item name
	 */
	public void renameCategoryItem(String catName, String oldItem, String newItem)
	{
		JCards.log("Renaming "+oldItem+" to "+newItem+" in category "+catName, 5);
		Vector catItems = (Vector)categories.get(catName);
		if (catItems != null)
		{
			int pos = catItems.indexOf(oldItem);
			catItems.removeElementAt(pos);
			catItems.insertElementAt(newItem, pos);
			setChanged();
			notifyObservers(CATEGORY_ITEM_RENAMED);
		}
	}

	/**
	 * Remove a category item
	 * @param catName category name
	 * @param item category item to remove
	 */
	public void removeCategoryItem(String catName, String item)
	{
		JCards.log("Removing "+item+" from category "+catName, 5);
		Vector catItems = (Vector)categories.get(catName);
		if (catItems != null)
		{
			catItems.removeElement(item);
			setChanged();
			notifyObservers(CATEGORY_ITEM_REMOVED);
		}
	}

	public void moveCategoryItem(String catName, String item, int delta)
	{
		JCards.log("Moving category item "+item+" "+delta, 5);
		Vector catItems = (Vector)categories.get(catName);
		if (catItems != null)
		{
			int pos = catItems.indexOf(item);
			catItems.removeElement(item);
			catItems.insertElementAt(item, pos+delta);
			setChanged();
			notifyObservers(CATEGORY_ITEM_MOVED);
		}
	}

	/**
	 * Test for whether any data definitions have been set up
	 * @return true if some definition has been set up
	 */
	public boolean hasDataBeenDefined()
	{
		if (fieldDefs.size() == 0 && categories.size() == 0) return false;
		else return true;
	}

	/**
	 * Set which fields are visible in the list. If no field names are selected
	 * then the first field is automatically set to be visible.
	 * @param visFields array of field names
	 */
	public void setVisibleFields(String[] visFields)
	{
		JCards.log("Setting visible fields", 5);

		// one field must always be visible
		if (visFields.length == 0)
		{
			visibleFields = new Vector(1);
			visibleFields.addElement(getFieldDef(0).getName());
		}
		else
		{
			visibleFields = new Vector(visFields.length);
			for (int i=0; i<visFields.length; i++)
			{
				visibleFields.addElement(visFields[i]);
			}
		}
		setChanged();
		notifyObservers(VISIBLE_FIELDS);
	}

	/**
	 * Get which fields are currently visible
	 * @return an array of visible fields or null
	 */
	public String[] getVisibleFieldsArray()
	{
		if (visibleFields == null) return null;
		else
		{
			String[] visFields = new String[visibleFields.size()];
			int index = 0;
			Enumeration enm = visibleFields.elements();
			while (enm.hasMoreElements())
			{
				visFields[index++] = (String)enm.nextElement();
			}

			return visFields;
		}
	}

	/**
	 * Get which fields are currently visible
	 * @return an array of visible fields or null
	 */
	public Vector getVisibleFields()
	{
		return visibleFields;
	}

	/**
	 * Is the today's date prepopulated in date fields
	 * @return true if prepopulation is on
	 */
	public boolean isDatePrepopulated()
	{
		return prepopulateDate;
	}

	/**
	 * Set whether to prepopulate date fields
	 * @param prepopulate set to true to prepopulate date fields
	 */
	public void setDatePrepopulated(boolean prepopulate)
	{
		prepopulateDate = prepopulate;
	}

	/**
	 * Set the date format. This cannot be null.
	 * @param newDateFormat the new date format
	 * @return false if newDateFormat is not legal
	 */
	public boolean setDateFormat(String newDateFormat)
	{
		if (newDateFormat == null || newDateFormat.length() == 0) return false;
		try {
			sdf.applyPattern(newDateFormat);
		} catch (IllegalArgumentException ex) {
			JCards.log("Incorrect date format: "+ex.getMessage(), 1);
			return false;
		}
		dateFormat = newDateFormat;
		JCards.log("Set date format to "+dateFormat, 9);
		return true;
	}

	/**
	 * Get the date format
	 * @return the current date format
	 */
	public String getDateFormat()
	{
		return dateFormat;
	}

	/**
	 * Convenience method for parsing a Date into a String which means that
	 * other classes don't have to keep setting up SimpleDateFormat objects.
	 * @param d the Date to format
	 * @return the Date parsed into a String using the current date format
	 */
	public String formatDate(Date d)
	{
		if (d == null) {
			return null;
		} else {
			return sdf.format(d);
		}
	}

	/**
	 * Convenience method for parsing a String into a Date which means that
	 * other classes don't have to keep setting up SimpleDateFormat objects.
	 * @param s the date to parse
	 * @return a Date object
	 * @exception ParseException thrown if the date cannot be parsed
	 */
	public Date parseDate(String s) throws ParseException
	{
		if (s == null) {
			return null;
		} else {
			return sdf.parse(s);
		}
	}

	/**
	 * Set the primary field to sort on
	 * @param sField the field to sort on
	 * @param direction the direction to sort in - Cards.SORT_ASCENDING or Cards.SORT_DESCENDING
	 */
	public void setSortField(String sField, int direction)
	{
		if (sField != null)
		{
			int oldDirection = sortDirection;
			String oldSort = sortField;
			sortDirection = direction;
			sortField = sField;
			secondSortField = null;

			if (oldDirection != direction || oldSort == null || !oldSort.equals(sField))
			{
				setChanged();
				notifyObservers(FIELD_SORT);
			}
		}
	}

	/**
	 * Set the secondary sort field
	 * @param ssField secondary sort field
	 */
	public void setSecondSortField(String ssField)
	{
		String oldSecond = secondSortField;
		secondSortField = ssField;

		if (oldSecond != null && ssField != null && !oldSecond.equals(ssField))
		{
			setChanged();
			notifyObservers(FIELD_SORT);
		}
	}

	/**
	 * Set the field(s) to sort on
	 * @param sField the field to sort on
	 * @param ssField the second field to sort on - may be null
	 * @param direction the direction to sort in - Cards.SORT_ASCENDING or Cards.SORT_DESCENDING
	 */
	public void setSortCriteria(String sField, String ssField, int direction)
	{
		if (sField != null)
		{
			int oldDirection = sortDirection;
			String oldSort = sortField;
			String oldSecond = secondSortField;
			sortDirection = direction;
			sortField = sField;
			secondSortField = ssField;

			// see if sort criteria has changed
			if ((oldDirection != direction) ||
				(oldSort == null || !oldSort.equals(sField)) ||
				(oldSecond != null && ssField != null && !oldSecond.equals(ssField)))
			{
				setChanged();
				notifyObservers(FIELD_SORT);
			}
		}
	}

	/** Get the sort field */
	public String getSortField()
	{
		return sortField;
	}

	/** Get the second sort field - may be null */
	public String getSecondSortField()
	{
		return secondSortField;
	}

	/**
	 * Get the sort direction
	 * @return sort direction
	 * @see Cards.SORT_ASCENDING
	 * @see Cards.SORT_DESCENDING
	 */
	public int getSortDirection()
	{
		return sortDirection;
	}

	/** Set whether field names should be displayed */
	public void setDisplayFieldNames(boolean dfn)
	{
		displayFieldNames = dfn;
	}

	/** Should field names be displayed */
	public boolean displayFieldNames()
	{
		return displayFieldNames;
	}

	/**
	 * Get an XML representation of this DBDef
	 * @return XMLElement of this DBDef
	 * @see nrs.jcards.io.DBDefXML
	 */
	public XMLElement getXML()
	{
		return DBDefXML.getXML(this);
	}

	/**
	 * Use XML to set this DBDef
	 * @param xmlImport the data to import
	 * @exception InvalidFieldTypeException unknown field type detected
	 * @exception InvalidFieldException field definition is incorrect
	 * @see nrs.jcards.io.DBDefXML
	 */
	public void setXML(XMLElement xmlImport) throws InvalidFieldTypeException, InvalidFieldException
	{
		// remove existing data
		reset();

		fieldDefs = new Vector();
		oldFieldDef = null;
		newFieldDef = null;

		DBDefXML.setXML(this, xmlImport);

		setChanged();
		notifyObservers(RELOAD);
	}
}
