package nrs.jcards;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import nrs.jcards.Card;
import nrs.jcards.dialog.CardDialog;
import nrs.jcards.dialog.InfoDialog;
import nrs.util.PostfixCalculator;

/*
 * This file is part of JCards.
 * JCards is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * JCards is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JCards; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * Class which displays and edits Cards
 *
 * @author Copyright &#169; 2002-2004 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class DisplayCardPanel extends Panel implements FocusListener
{
	private JCards jcards = null;
	private Hashtable entries = null;
	private String dateFormat = null;
	private boolean newCard = false;
	private boolean prepopulateDates = false;
	private boolean displayFieldNames = true;
	private DBDef def = null;
	private final static Font optionalLabelFont = new Font("SansSerif", Font.PLAIN, 12);
	private final static Font mandatoryLabelFont = new Font("SanSerif", Font.BOLD, 12);
	public final static String ERROR = "ERROR";
	public final static int ERROR_NUM = -99999;

	/**
	 * Initialise the DisplayCardPanel
	 * @param newCard set to true if editing a new card
	 */
	public DisplayCardPanel(JCards jcards, boolean newCard)
	{
		this.jcards = jcards;

		add(new Label("NO SELECTION"));

		this.newCard = newCard;

		// cache some of the settings from DBDef
		def = DBDef.getInstance();
		dateFormat = def.getDateFormat();
		prepopulateDates = def.isDatePrepopulated();
		displayFieldNames = def.displayFieldNames();

		setBackground(Color.white);
	}

	/**
	 * Display a card
	 * @param card the Card definition
	 * @param editing if true then editing is allowed
	 * @param wordWrap if true then word wrap memo fields
	 * @param cDialog the CardDialog so we can edit categories
	 */
	public void display(Card card, boolean editing, boolean wordWrap, CardDialog cDialog)
	{
		FieldDef fd;
		Label l;
		Component c;
		TextComponent tc;
		Checkbox cb;
		Choice ch;
		Boolean check;
		Button catButton;
		Object objValue = null;
		Date date;
		Vector catItems;
		Enumeration enm;
		String defaultValue = "";
		boolean displayFieldLabels = def.displayFieldNames();

		// don't bother if not displayed
		if (!isVisible()) return;

		int numberOfFields = def.getNumberOfFields();
		entries = new Hashtable(numberOfFields);

		// turn off visible updates
		setVisible(false);

		// remove any existing components
		removeAll();

		Panel fieldPanel = new Panel();
		GridBagLayout gbl = new GridBagLayout();
		fieldPanel.setLayout(gbl);
		GridBagConstraints gbc = new GridBagConstraints();

		gbc.insets = new Insets(2, 2, 2, 2);
		gbc.anchor = GridBagConstraints.NORTHWEST;
		gbc.weightx = 0.0;
		gbc.weighty = 0.0;

		// set up and optionally populate the fields
		// don't display the line if not editing and the field has no value
		for (int i=0; i<numberOfFields; i++)
		{
			fd = def.getFieldDef(i);
			objValue = null;
			if (card != null) {
				objValue = card.getValue(fd.getName());
			}

			// don't display lines with no data
			if (!editing && objValue == null) continue;

			// default value only applies to new cards
			if (newCard) {
				defaultValue = fd.getDefaultValue();
			}

			// add label
			if (displayFieldLabels || editing)
			{
				gbc.gridx = GridBagConstraints.RELATIVE;
				gbc.gridwidth = 1;
				gbc.fill = GridBagConstraints.NONE;
				gbc.weightx = 0.0;
				l = new Label(fd.getName());
				if (fd.isMandatory()) {
					l.setFont(mandatoryLabelFont);
				} else {
					l.setFont(optionalLabelFont);
				}
				gbl.setConstraints(l, gbc);
				fieldPanel.add(l);
			}

			// add input/display field
			switch (fd.getType())
			{
				case FieldDef.BOOLEAN:
					cb = new Checkbox();
					if (objValue != null)
					{
						check = (Boolean)objValue;
						if (check.booleanValue()) cb.setState(true);
					}
					cb.setEnabled(editing);
					if (defaultValue.equals(FieldDef.YES)) {
						cb.setState(true);
					}
					addComponent(fieldPanel, cb, gbl, gbc);
					if (editing) {
						entries.put(fd.getName(), cb);
					}
					break;

				case FieldDef.DATE:
					c = getDate(objValue, defaultValue, editing);
					addComponent(fieldPanel, c, gbl, gbc);
					if (editing) {
						entries.put(fd.getName(), c);
					}
					break;

				case FieldDef.MEMO:
					addMemo(objValue, editing, fieldPanel, gbl, gbc, fd, defaultValue, wordWrap);
					break;

				case FieldDef.CATEGORY:
					addCategory(objValue, editing, fieldPanel, gbl, gbc, fd, cDialog);
					break;

				case FieldDef.COUNTER:
					addCounter(objValue, editing, fieldPanel, gbl, gbc, fd);
					break;

				case FieldDef.CALCULATED:
					c = getCalculated(objValue);
					addComponent(fieldPanel, c, gbl, gbc);
					if (editing) {
						entries.put(fd.getName(), c);
					}
					break;

				default:
					c = getText(objValue, defaultValue, editing);
					addComponent(fieldPanel, c, gbl, gbc);
					if (editing) {
						entries.put(fd.getName(), c);
					}
					break;
			}
		}

		// add one line which takes up any extra vertical space
		gbc.weighty = 1.0;
		gbc.fill = GridBagConstraints.BOTH;
		l = new Label();
		gbl.setConstraints(l, gbc);
		fieldPanel.add(l);

		ScrollPane scroll = new ScrollPane();
		scroll.add(fieldPanel);

		setLayout(new BorderLayout());
		add(scroll, BorderLayout.CENTER);

		validate();

		// display all updates
		setVisible(true);
	}

	/** Set the layout proerties and add the component */
	private void addComponent(Panel p, Component c, GridBagLayout gbl, GridBagConstraints gbc)
	{
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbc.weightx = 1.0;
		gbl.setConstraints(c, gbc);
		p.add(c);
	}

	/** Get a date component */
	private Component getDate(Object val, String defaultValue, boolean editing)
	{
		Component c;

		if (editing)
		{
			TextField tc = new TextField(dateFormat);
			if (val != null) {
				tc.setText(def.formatDate((Date)val));
			}
			else
			{
				if (prepopulateDates) {
					tc.setText(def.formatDate((Date)val));
				} else if (defaultValue.length() > 0) {
					tc.setText(defaultValue);
				} else {
					tc.setText(dateFormat);
					tc.setForeground(Color.darkGray);
				}
			}
			tc.setEditable(editing);
			tc.addFocusListener(this);
			c = tc;
		}
		else
		{
			Label l = new Label(def.formatDate((Date)val));
			c = l;
		}

		return c;
	}

	/** Get a text Component */
	private Component getText(Object val, String defaultValue, boolean editing)
	{
		Component c;

		if (editing)
		{
			TextField tf = new TextField(10);
			tf.setEditable(editing);
			if (val != null) {
				tf.setText(val.toString());
			} else {
				tf.setText(defaultValue);
			}
			c = tf;
		}
		else
		{
			Label l = new Label();
			if (val != null) {
				l.setText(val.toString());
			}
			c = l;
		}

		return c;
	}

	/** Add a Category component */
	private void addCategory(Object val, boolean editing, Panel p, GridBagLayout gbl,
		GridBagConstraints gbc, FieldDef fd, CardDialog cDialog)
	{
		if (editing)
		{
			Choice ch = new Choice();
			// add a blank item first
			ch.add("");
			// add category item
			Vector catItems = DBDef.getInstance().getCategory(fd.getCategory());
			if (catItems != null)
			{
				Enumeration enm = catItems.elements();
				while (enm.hasMoreElements()) {
					ch.add((String)enm.nextElement());
				}
			}
			if (val != null) {
				ch.select((String)val);
			} else if (newCard) {
				ch.select(fd.getDefaultValue());
			}
			gbc.gridwidth = 1;
			gbl.setConstraints(ch, gbc);
			p.add(ch);
			entries.put(fd.getName(), ch);

			// add Button for editing
			Button catButton = new Button("...");
			catButton.setActionCommand("..."+fd.getName());
			catButton.addActionListener(cDialog);
			gbc.gridwidth = GridBagConstraints.REMAINDER;
			gbc.weightx = 0.0;
			gbl.setConstraints(catButton, gbc);
			p.add(catButton);
		}
		else
		{
			Label l = new Label((String)val);
			gbc.gridwidth = GridBagConstraints.REMAINDER;
			gbc.weightx = 1.0;
			gbc.fill = GridBagConstraints.HORIZONTAL;
			gbl.setConstraints(l, gbc);
			p.add(l);
		}
	}

	/** Add a Memo component */
	private void addMemo(Object val, boolean editing, Panel p, GridBagLayout gbl,
		GridBagConstraints gbc, FieldDef fd, String defaultValue, boolean wordWrap)
	{
			int memolines = fd.getMemoLines();
			int scrollPolicy = (wordWrap ? TextArea.SCROLLBARS_VERTICAL_ONLY :  TextArea.SCROLLBARS_BOTH);
			if (!editing && val != null) {
				// set number of lines equal to lines entered
				memolines = countNewlines((String)val);
			}

			TextArea tc = new TextArea("", memolines, 20, scrollPolicy);
			if (val != null) {
				tc.setText((String)val);
			} else {
				tc.setText(defaultValue);
			}
			tc.setEditable(editing);
			gbc.gridx = 0;
			gbc.gridwidth = GridBagConstraints.REMAINDER;
			gbc.fill = GridBagConstraints.BOTH;
			gbc.weightx = 1.0;
			gbc.weighty = 1.0;
			gbl.setConstraints(tc, gbc);
			p.add(tc);
			if (editing) {
				entries.put(fd.getName(), tc);
			}
	}

	/** Add a Counter component */
	private void addCounter(Object val, boolean editing, Panel p, GridBagLayout gbl,
		GridBagConstraints gbc, FieldDef fd)
	{
		Label cVal = new Label();

		if (editing && val == null && newCard)
		{
			// add counter item
			Counter count = DBDef.getInstance().getCounter(fd.getCounter());
			if (count != null)
			{
				count.increment();
				cVal.setText(Double.toString(count.getValue()));
			}
			else
			{
				cVal.setText(ERROR);
			}
		}
		else if (val != null)
		{
			cVal.setText(val.toString());
		}

		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbl.setConstraints(cVal, gbc);
		p.add(cVal);
		entries.put(fd.getName(), cVal);
	}

	/** Get a calculated field */
	private Component getCalculated(Object val)
	{
		Label cVal = new Label();
		if (val != null) {
			cVal.setText(val.toString());
		}

		return cVal;
	}

	/**
	 * Update the calculated fields.
	 * <p>Formulae are nominally correct as they are
	 * checked when entered, however due to field
	 * changes they may no longer be correct.
	 * @exception InvalidFieldException a field defined in the calculation no longer exists
	 */
	public void updateCalculations() throws InvalidFieldException
	{
		FieldDef fd;
		String calc, token;
		StringBuffer sb;
		StringTokenizer st;
		Label calcLabel;
		TextField tf;
		Component comp;
		PostfixCalculator calculator = DBDef.getCalculator();
		Double result;

		for (int i=0; i<def.getNumberOfFields(); i++)
		{
			fd = def.getFieldDef(i);
			if (fd.getType() == FieldDef.CALCULATED)
			{
				calc = fd.getCalculation();
				calcLabel = (Label)entries.get(fd.getName());
				sb = new StringBuffer(calc.length());
				st = new StringTokenizer(calc, " ");
				while (st.hasMoreTokens())
				{
					token = st.nextToken();
					if (token.startsWith("'"))
					{
						while (!token.endsWith("'"))
						{
							token += " ";
							token += st.nextToken();
						}
						// remove quotes
						token = token.substring(1, token.length()-1);

						// get field value
						comp = (Component)entries.get(token);
						if (comp == null) {
							calcLabel.setText(ERROR);
							throw new InvalidFieldException(token+" doesn't exist any more");
						}
						if (comp instanceof TextField) {
							tf = (TextField)comp;
							if (tf.getText().length() == 0) {
								sb.append("0");
							} else {
								sb.append(tf.getText());
							}
						} else if (comp instanceof Label) {
							sb.append(((Label)comp).getText());
						}
					}
					else
					{
						sb.append(token);
					}
					sb.append(" ");
				}

				// calculate the value
				try {
					calculator.parseCalculation(sb.toString());
					if (jcards.getLoggingLevel() >= 7) {
						System.out.println("Calculator stack contents");
						System.out.println(calculator);
					}
					result = calculator.getResult();
					calcLabel.setText(result.toString());
				} catch (IllegalArgumentException iex) {
					calcLabel.setText("ERROR");
					InfoDialog.showInfoDialog(jcards, jcards.getMessage("error"), InfoDialog.DIALOG_INFO, iex.getMessage());
				}
			}
		}
	}

	/** Rollback the counter fields */
	public void rollbackCounters()
	{
		FieldDef fd;
		Counter c;
		Vector fds = def.getFieldDefs();
		Enumeration enm = fds.elements();
		while (enm.hasMoreElements())
		{
			fd = (FieldDef)enm.nextElement();
			if (fd.getType() == FieldDef.COUNTER)
			{
				c = def.getCounter(fd.getCounter());
				c.rollback();
			}
		}
	}

	/**
	 * Get all the Components being edited.
	 * @return a Hashtable of Components
	 */
	public Hashtable getEntries()
	{
		return entries;
	}

	/**
	 * This method is called whenever a date field gets the focus.
	 * The text field will be set to blank if it contains the date hint.
	 */
	public void focusGained(FocusEvent e)
	{
		Component com = e.getComponent();
		TextField tf = (TextField)com;
		if (tf.getText().equals(dateFormat))
		{
			tf.setText(null);
			tf.setForeground(Color.black);
		}
	}

	/**
	 * This method is called whenever a date field loses the focus.
	 * If the text field is blank the date hint will be reapplied.
	 */
	public void focusLost(FocusEvent e)
	{
		TextField tf = (TextField)e.getComponent();

		if (tf.getText().length() == 0)
		{
			if (prepopulateDates) {
				tf.setText(def.formatDate(new Date()));
			} else {
				tf.setText(dateFormat);
			}
			tf.setForeground(Color.darkGray);
		}
	}

	/** Count the number of newlines in a String */
	private int countNewlines(String val)
	{
		StringTokenizer stz = new StringTokenizer(val, "\n");
		return stz.countTokens();
	}

	/**
	 * Iterate through memo fields setting caret position to 0.
	 * This cannot be done at initialisation because the AWT peer hasn't been created.
	 */
	public void setMemoStart()
	{
		if (!isShowing()) return;

		Object o;
		TextArea ta;
		Enumeration enm = entries.elements();
		while (enm.hasMoreElements())
		{
			o = enm.nextElement();
			if (o instanceof TextArea) {
				ta = (TextArea)o;
				ta.setCaretPosition(0);
			}
		}
	}
}
