package nrs.jcards;

import java.text.ParseException;
import java.util.Enumeration;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.Vector;
import nanoxml.XMLElement;
import nrs.jcards.io.Elements;


/*
 * This file is part of JCards.
 * JCards is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * JCards is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JCards; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * A field definition.
 *
 * @author Copyright &#169; 2002-2004 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class FieldDef
{
	/** Integer representations of the types */
	public final static int UNINIT = -1;
	public final static int STRING = 0;
	public final static int INTEGER = 1;
	public final static int DOUBLE = 2;
	public final static int DATE = 3;
	public final static int BOOLEAN = 4;
	public final static int MEMO = 5;
	public final static int CATEGORY = 6;
	public final static int COUNTER = 7;
	public final static int CALCULATED = 8;
	public final static int LAST_TYPE = CALCULATED;;

	/** String representations of the types */
	public final static String STRING_NAME = "String";
	public final static String INTEGER_NAME = "Integer";
	public final static String DOUBLE_NAME = "Double";
	public final static String DATE_NAME = "Date";
	public final static String BOOLEAN_NAME = "Boolean";
	public final static String MEMO_NAME = "Memo";
	public final static String CATEGORY_NAME = "Category";
	public final static String COUNTER_NAME = "Counter";
	public final static String CALCULATED_NAME = "Calculated";

	public final static String[] typeNames = { STRING_NAME, INTEGER_NAME,
		DOUBLE_NAME, DATE_NAME, BOOLEAN_NAME, MEMO_NAME, CATEGORY_NAME,
		COUNTER_NAME, CALCULATED_NAME };

	 private int type = UNINIT;
	 private String name = null;
	 private String catName = null;
	 private String countName = null;
	 private String defaultValue = "";
	 private boolean mandatory = false;
	 private int memoLines = 3;

	public static String YES = "Y";
	public static String NO = "N";

	static {
		ResourceBundle labels = JCards.getLabelsBundle();
		YES = labels.getString("initial.yes");
		NO = labels.getString("initial.no");
	}

	 public FieldDef()
	 {}

	 /**
	  * Create a new field definition
	  * @param type type of the field
	  * @param name name of the field
	  */
	 public FieldDef(int type, String name)
	 {
		 this.type = type;
		 this.name = name;
	 }

	 /** Create a new field definition by importing from XML */
	 public FieldDef(XMLElement fdXML) throws InvalidFieldTypeException
	 {
		 String attrName;

		 Enumeration enm = fdXML.enumeratePropertyNames();
		 while (enm.hasMoreElements())
		 {
			 attrName = (String)enm.nextElement();

			 if (attrName.equalsIgnoreCase(Elements.NAME))
			 {
				 name = fdXML.getProperty(Elements.NAME);
			 }
			 else if (attrName.equalsIgnoreCase(Elements.TYPE))
			 {
				 setType(fdXML.getProperty(Elements.TYPE));
			 }
			 else if (attrName.equalsIgnoreCase(Elements.MANDATORY))
			 {
				 Boolean b = new Boolean(fdXML.getProperty(Elements.MANDATORY));
				 mandatory = b.booleanValue();
			 }
			 else if (attrName.equalsIgnoreCase(Elements.FIELD_CAT))
			 {
				 catName = fdXML.getProperty(Elements.FIELD_CAT);
			 }
			 else if (attrName.equalsIgnoreCase(Elements.COUNTER))
			 {
				 countName = fdXML.getProperty(Elements.COUNTER);
			 }
			 else if (attrName.equalsIgnoreCase(Elements.MEMO_LINES))
			 {
				 memoLines = Integer.parseInt(fdXML.getProperty(Elements.MEMO_LINES));
			 }
			 else if (attrName.equalsIgnoreCase(Elements.DEFAULT))
			 {
				 defaultValue = fdXML.getProperty(Elements.DEFAULT);
			 }
			 else if (attrName.equalsIgnoreCase(Elements.CALC))
			 {
				 defaultValue = fdXML.getProperty(Elements.CALC);
			 }
		 }
	 }

	 /** This field uses a category to get its values */
	 public void setCategory(String newCat)
	 {
		 catName = newCat;
	 }

	 /**
	  * Get the field's category
	  * @return the category or null if not set
	  */
	 public String getCategory()
	 {
		 return catName;
	 }

	 /** This field uses a counter to get its values */
	 public void setCounter(String newCount)
	 {
		 countName = newCount;
	 }

	 /**
	  * Get the field's counter
	  * @return the counter or null if not set
	  */
	 public String getCounter()
	 {
		return countName;
	 }

	 /**
	  * Set the default value which is limited by the type,
	  * i.e. you can't set it to 'A' if the type is Integer
	  * @param value the value to set
	  * @return true if the default value has been set; false if it is of the incorrect type
	  */
	 public boolean setDefaultValue(String value)
	 {
		 if (value == null) {
			 value = "";
			 return true;
		 }

		 try
		 {
			 switch (type)
			 {
				 case STRING:
				 case MEMO:
				 case CALCULATED:
					 defaultValue = value;
					 break;

				 case DATE:
					 DBDef.getInstance().parseDate(value);
					 defaultValue = value;
					 break;

				 case INTEGER:
					 Integer.parseInt(value);
					 defaultValue = value;
					 break;

				 case DOUBLE:
					 Double.valueOf(value);
					 defaultValue = value;
					 break;

				 case CATEGORY:
					 if (catName == null) return false;
					 Vector cats = DBDef.getInstance().getCategory(catName);
					 if (cats.contains(value)) {
						 defaultValue = value;
						 break;
					 } else return false;

				 case BOOLEAN:
					 String uc = value.toUpperCase();
					 if (uc.equals(YES) || uc.equals("Y") || uc.equals("1") || uc.equals("true")) {
						 defaultValue = YES;
					 } else if (uc.equals(NO) || uc.equals("N") || uc.equals("0") || uc.equals("false")) {
						 defaultValue = NO;
					 }
					 break;

				 default:
					 return false;
			 }
		 } catch (Exception ex) {
			 return false;
		 }

		 return true;
	 }

	 /** Get the default value */
	 public String getDefaultValue()
	 {
		 return defaultValue;
	 }

	 /**
	  * Synonym for getDefaultValue
	  * @see #getDefaultValue
	  */
	 public String getCalculation()
	 {
		 return getDefaultValue();
	 }

	 /**
	  * Get the type of this field definition
	  */
	 public int getType()
	 {
		 return type;
	 }

	 /**
	  * Get an array of Strings representing the available types
	  * @return
	  */
	 public static String[] getTypes()
	 {
		 return typeNames;
	 }

	 /**
	  * Set the type of the field to the specified value
	  * @param type type name
	  * @exception InvalidFieldTypeException if the field type is unknown
	  */
	 public void setType(String typeName) throws InvalidFieldTypeException
	 {
		 for (int i=0; i<typeNames.length; i++)
		 {
			 if (typeNames[i].equals(typeName))
			 {
				 type = i;
				 break;
			 }
		 }

		 if (type == UNINIT) throw new InvalidFieldTypeException("Field type "+type+" is not defined");
	 }

	 /**
	  * Return the number of types
	  * @return the number of field types
	  */
	 public static int getNumberOfTypes()
	 {
		 return LAST_TYPE;
	 }

	 /**
	  * Get the type as a name
	  * @param t the type to convert
	  * @return the converted name or null if it is not defined
	  */
	 public static String getTypeName(int t)
	 {
		 if (t>=0 && t<=LAST_TYPE)
			 return typeNames[t];
		 else
			 return null;
	 }

	 /**
	  * Convert the type name into an int
	  * @param typeName name of the type
	  * @return int representing the type
	  */
	 public static int getType(String typeName)
	 {
		 for (int i=0; i<typeNames.length; i++)
		 {
			 if (typeNames[i].equals(typeName)) {
				 return i;
			 }
		 }

		 return UNINIT;
	 }

	 /**
	  * Get the name of the field
	  * @return name of the field
	  */
	 public String getName()
	 {
		 return name;
	 }

	 /**
	  * Set the name of the field
	  * @param newName the name of the field
	  */
	 public void setName(String newName)
	 {
		 name = newName;
	 }

	 /**
	  * Set this field to be mandatory or not
	  * @param mand set to true if you want the field to be mandatory
	  */
	 public void setMandatory(boolean mand)
	 {
		 mandatory = mand;
	 }

	 /**
	  * Is this field mandatory
	  * @return true if mandatory
	  */
	 public boolean isMandatory()
	 {
		 return mandatory;
	 }

	 /** Set the number of lines used for this memo field */
	 public void setMemoLines(int ml)
	 {
		 memoLines = ml;
	 }

	 /** Get the number of lines used for a memo field */
	 public int getMemoLines()
	 {
		 return memoLines;
	 }

	 /**
	  * Based on the field type, return an Object of the right type.
	  * @param type field type
	  * @param value of the field
	  * @return an object representing the field value,
	  * or null if an unknown field type
	  * @exception NumberFormatException if a field is a number but the value is not
	  * @exception ParseException if the field is a date but the value is not
	  */
	 public static Object parseValue(int type, String value) throws NumberFormatException, ParseException
	 {
		 if (value == null || value.length() == 0) {
			return null;
		 }

		 switch (type)
		 {
			case STRING:
			case MEMO:
			case CATEGORY:
			case CALCULATED:
				return value;

			case INTEGER:
				return new Integer(value);

			case DOUBLE:
			case COUNTER:
				return new Double(value);

			case BOOLEAN:
				return new Boolean(value);

			case DATE:
				return DBDef.getInstance().parseDate(value);
		 }

		 return null;
	 }

	 /**
	  * Get a String representation of this class
	  * @return a String representation of this class
	  */
	 public String toString()
	 {
		 StringBuffer sb = new StringBuffer(name);
		 sb.append(",");
		 sb.append(getTypeName(type));
		 if (type != BOOLEAN)
		 {
			 sb.append(",");
			 sb.append(mandatory);
		 }
		 if (catName != null)
		 {
			 sb.append(",");
			 sb.append(catName);
		 }
		 if (countName != null)
		 {
			 sb.append(",");
			 sb.append(countName);
		 }

		 return sb.toString();
	 }

	 /**
	  * Get a representation of the data as XML
	  * @return XMLElement
	  */
	 public XMLElement getXML()
	 {
		 XMLElement fieldEl = new XMLElement();
		 fieldEl.setTagName(Elements.FIELD);
		 fieldEl.addProperty(Elements.NAME, name);
		 fieldEl.addProperty(Elements.TYPE, getTypeName(type));
		 fieldEl.addProperty(Elements.MANDATORY, new Boolean(mandatory));
		 if (catName != null) fieldEl.addProperty(Elements.FIELD_CAT, catName);
		 if (countName != null) fieldEl.addProperty(Elements.COUNTER, countName);
		 if (type == MEMO) fieldEl.addProperty(Elements.MEMO_LINES, memoLines);
		 if (defaultValue != null && defaultValue.length() > 0)
		 {
			 if (type == CALCULATED) {
				 fieldEl.addProperty(Elements.CALC, defaultValue);
			 } else {
				fieldEl.addProperty(Elements.DEFAULT, defaultValue);
			 }
		 }

		 return fieldEl;
	 }
}