package nrs.jcards;

import java.awt.*;
import java.awt.datatransfer.Clipboard;
import java.awt.event.*;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URL;
import java.text.MessageFormat;
import java.util.*;
import jclass.bwt.BWTEnum;
import jclass.bwt.JCActionEvent;
import jclass.bwt.JCActionListener;
import jclass.bwt.JCButton;
import nanoxml.XMLElement;
import nrs.jcards.dialog.*;
import nrs.jcards.io.JCAReader;

/*
 * JCards
 * Copyright (C) 2002-2005 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * Main application class for JCards.
 * See http://www.nsydenham.net/java/JCards/ for details.
 *
 * @author Copyright &#169; 2002-2005 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class JCards extends Frame implements ActionListener, ItemListener, JCActionListener, Observer, WindowListener
{
	public final static String VERSION = "5.1";

	private DBDef def = DBDef.getInstance();
	private Cards cards;
	private CardsList cardList = null;
	public static UserSettings settings = new UserSettings();
	private File cardFile = null;
	private Panel toolbar = null;
	private Menu viewMenu;
	private Menu lastMenu;
	private String prevSearchText = null;
	private boolean prevSearchMatchCase = false;
	private String prevSearchField = null;
	private final static int MAX_LAST_FILES = 5;
	private String[] lastFiles = new String[MAX_LAST_FILES];
	private transient String password = "";
	private static int logLevel = 0;
	private Dimension restoreSize = null;
	private long creationDate = 0;
	private long modifiedDate = 0;
	private final static ResourceBundle msgs = ResourceBundle.getBundle("nrs.jcards.props.msgs");

	// app settings
	public boolean changes = false;
	public static Font baseFont = new Font("SansSerif", Font.PLAIN, 12);
	private boolean viewToolbar = true;

	/** Preferences */
	private boolean saveConfirmation = true;
	private boolean openLastFile = true;
	private boolean autoEdit = true;
	private boolean wordWrap = true;

	/** Properties file */
	private final static String PROPS_DIR = ".jca";
	private final static String PROPS_FILE= "jca.properties";
	private final static String PROPS_SETTING = "jca.props";
	private final static String PROPS_OVERRIDE = "jcards.properties";

	/** Preference names */
	private final static String PREF_AUTO_EDIT = "autoEdit";
	private final static String PREF_FONT = "font";
	private final static String PREF_OPEN_LAST_FILE = "openLastFile";
	private final static String PREF_SAVE_CONF = "saveConfirmation";
	private final static String PREF_TOOLBAR = "toolbar";
	private final static String PREF_WORD_WRAP = "wordWrap";

	/** Only used for testing purposes */
	public JCards(boolean noGUI) {}

	/** Initialise the application GUI */
	public JCards()
	{
		super("JCards");

		// when user clicks close window decoration
		addWindowListener(this);

		// set up the list
		cards = new Cards(this);
		cardList = new CardsList(this);
		cards.addObserver(cardList);
		cards.addObserver(this);
		// if the DB definition changes the Cards
		// object might need to edit the data
		// We also need to know about changes
		def.addObserver(cards);
		def.addObserver(this);

		// set up last opened menu
		ResourceBundle labels = getMenuBundle();
		lastMenu = new Menu(labels.getString("open.recent"));

		setLayout(new BorderLayout());
		add(cardList, BorderLayout.CENTER);

		changes = false;
	}

	/**
	 * Log a message
	 * @param message message to log
	 * @param level the message won't appear unless the level >= application logging level
	 */
	public static void log(String message, int level)
	{
		if (level <= logLevel) {
			System.out.println(message);
		}
	}

	/** Get the currect logging level */
	public static int getLoggingLevel()
	{
		return logLevel;
	}

	/**
	 * Start the application
	 * @param args command-line arguments
	 */
	public static void main(String[] args)
	{
		// see if any properties have been set
		String props = System.getProperty(PROPS_SETTING);
		if (props != null)
		{
			File propsFile = new File(props, PROPS_OVERRIDE);
			Properties jprops = new Properties();
			try {
				jprops.load(new FileInputStream(propsFile));
			} catch (Exception ex) {
				System.err.println("Properties file not found at "+propsFile.getAbsolutePath());
				return;
			}

			String log = jprops.getProperty("jca.log", "0");
			try {
				logLevel = Integer.parseInt(log);
				if (logLevel < 0) logLevel = 0;
				else if (logLevel > 9) logLevel = 9;
			} catch (NumberFormatException nfex) {
				log("Log level is not a number", 0);
			}
			log("Logging level set to "+logLevel, 3);

			String pdaStr = jprops.getProperty("jca.pda", "false");
			log("jca.pda="+pdaStr, 3);
			String width = jprops.getProperty("jca.width");
			log("jca.width="+width, 3);
			String height = jprops.getProperty("jca.height");
			log("jca.height="+height, 3);
			if (pdaStr.equalsIgnoreCase("true")) {
				settings.setPda(true);
				log("PDA flag set", 3);
			}
			try
			{
				if (width != null) settings.setOverrideWidth(Integer.parseInt(width));
				if (height != null) settings.setOverrideHeight(Integer.parseInt(height));
			} catch (NumberFormatException nfex) {
				log("Invalid width/height", 0);
			}
		}

		// see if a file was specified on the command line
		String cmdLineFile = null;
		if (args.length > 0)
		{
			log("File specified on command line", 3);
			File f = new File(args[0]);
			if (f.exists()) {
				cmdLineFile = f.getAbsolutePath();
			} else {
				System.err.println(args[0]+" doesn't exist!");
			}
		}

		// initialise the app
		JCards jcards = new JCards();
		String lastFile = jcards.readProperties();
		jcards.addToolbar();
		jcards.pack();
		jcards.setInitialLocation();
		// delay initialization of the menubar to get round a bug
		// in Insignia's Jeode JVM which is used in PDAs
		jcards.createMenuBar();
		jcards.show();

		// attempt to open command line file if specified
		if (cmdLineFile != null)
		{
			jcards.log("Reading command line file: "+cmdLineFile, 6);
			try {
				jcards.openFile(cmdLineFile);
			} catch (Exception ex) {
				MessageFormat msgFormat = new MessageFormat(jcards.getMessage("cannot.open.file"));
				InfoDialog.showInfoDialog(jcards, jcards.getMessage("error"), InfoDialog.DIALOG_INFO, msgFormat.format(cmdLineFile));
			}
		}
		else if (lastFile != null)
		{
			jcards.log("Reading last opened file: "+lastFile, 6);
			// attempt to read last opened file
			File f = new File(lastFile);
			if (f.exists())	{
				try {
					jcards.openFile(lastFile);
				} catch (IOException ioex) {
					System.err.println("Cannot open last file: "+ioex.getMessage());
				}
			} else {
				jcards.log("Last opened file doesn't exist", 2);
			}
		}
	}

	/** Set the initial location of the window */
	public void setInitialLocation()
	{
		if (settings.isPda())
		{
			setLocation(0, 0);
		}
		else
		{
			if (settings.getInitLocationX() > -1) {
				setLocation(settings.getInitLocationX(), settings.getInitLocationY());
			} else {
				centreWindow(this, this);
			}
		}
	}

	/**
	 * Add menu items
	 * @param menu Menu object to add the items to
	 * @param miObjs array of Strings, MenuShortcuts and other Menus
	 * @param labels ResourceBundle for the menu labels
	 */
	private void addMenuItems(Menu menu, Object[] miObjs, ResourceBundle labels)
	{
		Object miObj;
		String miName, miKey;
		MenuItem mi = null;

		for (int i=0; i<miObjs.length; i++)
		{
			miObj = miObjs[i];
			if (miObj instanceof String)
			{
				miName = (String)miObjs[i];
				if (miName.length() == 0) {
					menu.addSeparator();
				}
				else
				{
					boolean elipsis = false;
					if (miName.endsWith("...")) {
						elipsis = true;
						miKey = miName.substring(0, miName.length()-3);
					} else {
						miKey = miName;
					}
					miName = labels.getString(miKey);
					if (elipsis) {
						miName += "...";
					}
					mi = new MenuItem(miName);
					mi.setActionCommand(miKey);
					mi.addActionListener(this);
					menu.add(mi);
				}
			}
			else if (miObj instanceof MenuShortcut)
			{
				if (!settings.isPda()) {
					mi.setShortcut((MenuShortcut)miObj);
				}
			}
			else if (miObj instanceof Menu)
			{
				menu.add((Menu)miObj);
			}
		}
	}

	/**
	 * Create the menus. The MenuBar is passed in to try and get
	 * around a bug in Insignia's Jeode JVM.
	 * @return the created MenuBar
	 */
	private void createMenuBar()
	{
		int i;
		String miName;
		MenuItem mi;
		ResourceBundle labels = getMenuBundle();
		MenuBar bar = new MenuBar();

		// file menu
		Menu fileMenu = new Menu(labels.getString("file"));
		Object[] fileMenuItems = {
			"new", new MenuShortcut(KeyEvent.VK_N), "open...", new MenuShortcut(KeyEvent.VK_O),
			lastMenu, "", "save", new MenuShortcut(KeyEvent.VK_S), "save.as...", new MenuShortcut(KeyEvent.VK_A),
			"pass.set...", new MenuShortcut(KeyEvent.VK_P, true), "pass.clear", "", "csv.import...", "csv.export...", "cat.export...", "",
			"exit", new MenuShortcut(KeyEvent.VK_E) };
		addMenuItems(fileMenu, fileMenuItems, labels);
		bar.add(fileMenu);
		fileMenuItems = null;

		// Edit menu
		Menu editMenu = new Menu(labels.getString("edit"));
		Object[] editMenuItems = {
			"card.new", new MenuShortcut(KeyEvent.VK_N, true), "card.edit", new MenuShortcut(KeyEvent.VK_E, true),
			"card.del", new MenuShortcut(KeyEvent.VK_DELETE), "card.undel", new MenuShortcut(KeyEvent.VK_Z), "",
			"card.copy", new MenuShortcut(KeyEvent.VK_C), "copy.clip", new MenuShortcut(KeyEvent.VK_B), "",
			"find...", new MenuShortcut(KeyEvent.VK_F), "find.reset", new MenuShortcut(KeyEvent.VK_R) };
		addMenuItems(editMenu, editMenuItems, labels);
		bar.add(editMenu);
		editMenuItems = null;

		bar.add(editMenu);

		// DB menu
		Menu dbMenu = new Menu(labels.getString("db"));

		Object[] dbMenuItems = {
			"fields...", new MenuShortcut(KeyEvent.VK_F, true), "categories...", new MenuShortcut(KeyEvent.VK_C, true),
			"counters...", "comments...", "", "db.stats", "field.stats...", "", "prefs...", new MenuShortcut(KeyEvent.VK_P) };
		addMenuItems(dbMenu, dbMenuItems, labels);
		bar.add(dbMenu);
		dbMenuItems = null;

		viewMenu = new Menu(labels.getString("view"));
		bar.add(viewMenu);

		Menu helpMenu = new Menu(labels.getString("help"));
		MenuItem about = new MenuItem(labels.getString("about"));
		about.setActionCommand("about");
		about.addActionListener(this);
		helpMenu.add(about);

		bar.add(helpMenu);

		setMenuBar(bar);
	}

	/** Create a toolbar */
	public void resetToolbar()
	{
		if (toolbar == null) {
			toolbar = new Panel(new FlowLayout(FlowLayout.LEFT, 1, 0));
		} else {
			toolbar.removeAll();
		}

		JCButton button;
		I18NIcon icon;
		Vector buttons = settings.getUserIcons();
		Enumeration enm = buttons.elements();
		while (enm.hasMoreElements())
		{
			icon = (I18NIcon)enm.nextElement();
			button = new JCButton(null, getImage(icon.getGraphic()), BWTEnum.STRING_LEFT);
			button.setActionCommand(icon.getKey());
			button.addActionListener(this);
			toolbar.add(button);
		}
		toolbar.validate();
	}

	/** Add a toolbar to the top of the app */
	private void addToolbar()
	{
		if (viewToolbar)
		{
			resetToolbar();
			add(toolbar, BorderLayout.NORTH);
			validate();
		}
	}

	/** Remove the toolbar from the app */
	private void removeToolbar()
	{
		if (toolbar != null) {
			remove(toolbar);
			validate();
		}
		toolbar = null;
	}

	/** JCButton action.
	 * @see #actionPerformed(ActionEvent)
	 */
	public void actionPerformed(JCActionEvent e)
	{
		ActionEvent ae = new ActionEvent(e.getSource(), e.getID(), e.getActionCommand());
		actionPerformed(ae);
	}

	/** Handle menu events */
	public void actionPerformed(ActionEvent e)
	{
		String ac = e.getActionCommand();
		log("ActionCommand: "+ac, 6);

		// file menu
		if (ac.equals("new")) newFile();
		else if (ac.equals("open"))
		{
			try	{
				openFile(null);
			} catch (Exception ex) {
				String lines[] = { getMessage("error.open"), ex.getMessage() };
				InfoDialog.showInfoDialog(this, getMessage("cannot.open"), InfoDialog.DIALOG_INFO, lines);
				cardFile = null;
			}
		}
		else if (ac.equals("save")) saveFile();
		else if (ac.equals("save.as")) saveFileAs();
		else if (ac.equals("pass.set")) enterPassword();
		else if (ac.equals("pass.clear")) clearPassword();
		else if (ac.equals("csv.import")) importCSV();
		else if (ac.equals("csv.export")) exportCSV();
		else if (ac.equals("cat.export")) exportCategories();
		else if (ac.equals("exit")) exit();

		// edit menu
		else if (ac.equals("card.copy")) editCard(false, true, true);
		else if (ac.equals("copy.clip")) copyToClipboard();
		else if (ac.equals("find")) find();
		else if (ac.equals("find.reset")) cards.resetFind();

		// db menu
		else if (ac.startsWith("card.new")) editCard(true, true, false);
		else if (ac.equals("card.edit")) editCard(false, autoEdit, false);
		else if (ac.startsWith("card.del")) deleteCard();
		else if (ac.equals("card.undel")) undeleteCard();

		else if (ac.equals("fields")) editFields();
		else if (ac.equals("categories")) editCategories();
		else if (ac.equals("counters")) editCounters();
		else if (ac.equals("comments")) editComment();

		else if (ac.equals("db.stats")) showStatistics(null);
		else if (ac.equals("field.stats")) calculateStatistics();

		else if (ac.equals("prefs")) editPrefs(false);

		else if (ac.equals("full.screen")) fullScreen();
		else if (ac.equals("toolbar")) itemStateChanged(null);

		// help menu
		else if (ac.equals("about")) showAbout();

		// open recent file
		else if (ac.startsWith("LF:"))
		{
			String file = ac.substring(3);
			try {
				openFile(file);
			} catch (Exception ex) {
				InfoDialog.showInfoDialog(this, getMessage("error.open"), InfoDialog.DIALOG_INFO, ex.getMessage());
			}
		}

		// View menu
		else if (ac.startsWith("CAT:"))
		{
			String catitem = ac.substring(4);
			String field = catitem.substring(0, catitem.indexOf(";"));
			String item = catitem.substring(catitem.indexOf(";")+1);
			findCards(item, field, true, false);
		}
	}

	/** User changed the toolbar checkbox menu item */
	public void itemStateChanged(ItemEvent e)
	{
		if (viewToolbar) {
			viewToolbar = false;
			removeToolbar();
		} else {
			viewToolbar = true;
			addToolbar();
		}
	}

	/**
	 * Check to see if any changes have been made
	 * @return true if the user Cancels the operation
	 */
	private boolean checkForChanges()
	{
		if (changes)
		{
			int response = InfoDialog.showInfoDialog(this, getMessage("save.changes"), InfoDialog.DIALOG_CONFIRM, getMessage("data.save"));
			switch (response)
			{
				case InfoDialog.YES:
					saveFile();
					return false;

				case InfoDialog.NO:
					return false;

				default:
					return true;
			}
		}

		return false;
	}

	// launch the find dialog
	private void find()
	{
		FindDialog findDialog = new FindDialog(this, prevSearchText, prevSearchField, prevSearchMatchCase);
		findDialog.pack();
		centreWindow(this, findDialog);
		findDialog.show();
	}

	// start a new file
	private void newFile()
	{
		if (checkForChanges()) return;

		reset();
	}

	/** Reset the app to defaults */
	private void reset()
	{
		cards.reset();
		cardList.reset();
		def.reset();
		cardFile = null;
		changes = false;
		password = "";
		prevSearchText = null;
		prevSearchMatchCase = false;
		prevSearchField = null;
		creationDate = (new Date()).getTime();

		setTitle("JCards");
	}

	/**
	 * Load a card file
	 * @param loadFile the file to load. If null a dialog box will open.
	 * @param IOException thrown if there is an error loading the file
	 */
	private void openFile(String loadFile) throws IOException
	{
		XMLElement el;
		String tagName;
		File fileToOpen;

		if (checkForChanges()) return;

		if (loadFile == null)
		{
			FileDialog openDialog = new FileDialog(this, getMessage("open.file"), FileDialog.LOAD);
			String dir;
			if (cardFile != null) {
				dir = cardFile.getParent();
			} else {
				dir = System.getProperty("user.home");
			}
			openDialog.setDirectory(dir);
			openDialog.show();
			String openFile = openDialog.getFile();
			dir = openDialog.getDirectory();

			if (openFile == null) {
				return;
			} else {
				fileToOpen = new File(dir, openFile);
			}
		} else {
			fileToOpen = new File(loadFile);
			// check it exists
			if (!fileToOpen.exists())
			{
				InfoDialog.showInfoDialog(this, getMessage("error"), InfoDialog.DIALOG_INFO, getMessage("no.file"));
				return;
			}
		}
		log("Loading file "+fileToOpen.getAbsolutePath(), 3);

		// reset all data structures
		reset();

		cardFile = fileToOpen;

		try
		{
			JCAReader reader = new JCAReader(this);
			reader.openFile(cardFile);
			setTitle("JCards:"+cardFile.getName());
			changes = false;
		} catch (Exception ex) {
			InfoDialog.showInfoDialog(this, getMessage("error"), InfoDialog.DIALOG_INFO, getMessage("errorfile"));
			System.err.println(ex.getMessage());
			ex.printStackTrace();
			cardFile = null;
		} finally {
			updateLastOpenedMenu(cardFile.getAbsolutePath());
			updateViewMenu();
		}

		log("Finished opening file", 5);
	}

	/** Update the View menu */
	private void updateViewMenu()
	{
		String field, cat, item;
		Vector catItems;
		Menu catMenu;
		MenuItem cMI;
		Enumeration cEnum, iEnum;
		FieldDef fd;
		ResourceBundle labels = getMenuBundle();

		if (viewMenu.getItemCount() > 0) {
			viewMenu.removeAll();
		}

		cMI = new MenuItem(labels.getString("full.screen"));
		cMI.setActionCommand("full.screen");
		cMI.addActionListener(this);
		viewMenu.add(cMI);

		CheckboxMenuItem cbMI = new CheckboxMenuItem(labels.getString("toolbar"), viewToolbar);
		cbMI.setActionCommand("toolbar");
		cbMI.addItemListener(this);
		viewMenu.add(cbMI);

		viewMenu.addSeparator();

		cMI = new MenuItem(labels.getString("all"));
		cMI.setActionCommand("find.reset");
		cMI.addActionListener(this);
		viewMenu.add(cMI);
		viewMenu.addSeparator();

		Vector fields = def.getFieldDefs();
		cEnum = fields.elements();
		while (cEnum.hasMoreElements())
		{
			fd = (FieldDef)cEnum.nextElement();
			if (fd.getType() == FieldDef.CATEGORY)
			{
				cat = fd.getCategory();
				field = fd.getName();
				catMenu = new Menu(field);
				catItems = def.getCategory(cat);
				iEnum = catItems.elements();
				while (iEnum.hasMoreElements())
				{
					item = (String)iEnum.nextElement();
					cMI = new MenuItem(item);
					cMI.setActionCommand("CAT:"+field+";"+item);
					cMI.addActionListener(this);
					catMenu.add(cMI);
				}
				viewMenu.add(catMenu);
			}
		}
	}

	/**
	 * Update the last opened menu
	 * @param file the last opened file
	 */
	private void updateLastOpenedMenu(String file)
	{
		int i, j;

		// see if file is already in last opened list
		for (i=0; i<MAX_LAST_FILES; i++)
		{
			if (lastFiles[i] != null && file.equals(lastFiles[i]))
			{
				// remove entry and move other's up
				for (j=i+1; j<MAX_LAST_FILES; j++)
				{
					lastFiles[i] = lastFiles[j];
				}
				break;
			}
		}

		// add file to last opened list
		for (i=MAX_LAST_FILES-1; i>0; i--)
		{
			lastFiles[i] = lastFiles[i-1];
		}
		lastFiles[0] = file;

		// update menu
		lastMenu.removeAll();
		MenuItem lfMI;
		File f;
		for (i=0; i<MAX_LAST_FILES; i++)
		{
			if (lastFiles[i] != null)
			{
				f = new File(lastFiles[i]);
				lfMI = new MenuItem(f.getName());
				lfMI.setActionCommand("LF:"+lastFiles[i]);
				lfMI.addActionListener(this);
				lastMenu.add(lfMI);
			}
		}
	}

	/** Save the data to an existing or new file */
	private void saveFile()
	{
		log("Save", 9);

		modifiedDate = (new Date()).getTime();
		if (creationDate == 0) {
			creationDate = modifiedDate;
		}

		if (cards.size() == 0 && !changes)
		{
			InfoDialog.showInfoDialog(this, getMessage("no.data"), InfoDialog.DIALOG_INFO, getMessage("no.data.save"));
			return;
		}

		if (cardFile == null)
		{
			log("No file specified", 5);
			FileDialog saveDialog = new FileDialog(this, getMessage("save.to"), FileDialog.SAVE);
			saveDialog.setDirectory(System.getProperty("user.home"));
			saveDialog.setFile("jcards.jca");
			saveDialog.show();
			String saveFile = saveDialog.getFile();
			String dir = saveDialog.getDirectory();
			saveDialog.dispose();

			if (saveFile == null) {
				return;
			} else {
				cardFile = new File(dir, saveFile);
			}
		}
		log("File to save to: "+cardFile.getAbsolutePath(), 5);

		MessageFormat msgFormat = new MessageFormat(getMessage("saved.to"));
		try
		{
			JCAReader reader = new JCAReader(this);
			reader.saveFile(cardFile, password);

			changes = false;

			if (saveConfirmation)
			{
				String[] msgLines;
				String[] fargs = { cardFile.getName() };
				if (password.length() > 0)
				{
					msgLines = new String[2];
					msgLines[0] = msgFormat.format(fargs);
					msgLines[1] = getMessage("file.enc");
				}
				else
				{
					msgLines = new String[1];
					msgLines[0] = msgFormat.format(fargs);
				}
				InfoDialog.showInfoDialog(this, getMessage("saved"), InfoDialog.DIALOG_INFO, msgLines);
			}
			updateLastOpenedMenu(cardFile.getAbsolutePath());
		} catch (Exception ex) {
			msgFormat = new MessageFormat(getMessage("cannot.save"));
			String msg = (ex.getMessage() == null ? getMessage("unknown.error") : ex.getMessage());
			String[] msgLines = { msgFormat.format(cardFile.getName()), ex.getMessage() };
			InfoDialog.showInfoDialog(this, getMessage("error"), InfoDialog.DIALOG_INFO, msgLines);
		}

		setTitle("JCards:"+cardFile.getName());
		log("Finished save", 9);
	}

	/** Save data to a new file */
	private void saveFileAs()
	{
		log("Save file as", 9);

		if (cards.size() == 0 && def.getNumberOfFields() == 0)
		{
			InfoDialog.showInfoDialog(this, getMessage("error"), InfoDialog.DIALOG_INFO, getMessage("no.data.save"));
			return;
		}

		FileDialog saveDialog = new FileDialog(this, getMessage("save.to"), FileDialog.SAVE);
		if (cardFile != null) {
			saveDialog.setDirectory(cardFile.getPath());
			saveDialog.setFile(cardFile.getName());
		} else {
			saveDialog.setDirectory(System.getProperty("user.home"));
			saveDialog.setFile("jcards.jca");
		}
		saveDialog.show();
		String newCardFile = saveDialog.getFile();
		saveDialog.dispose();

		if (newCardFile != null)
		{
			cardFile = new File(saveDialog.getDirectory(), newCardFile);
			saveFile();
		}

		log("Finished save file as", 9);
	}

	/** Export the data to a CSV file */
	void exportCSV()
	{
		if (cards.size() == 0 && def.getNumberOfFields() == 0)
		{
			InfoDialog.showInfoDialog(this, getMessage("error"), InfoDialog.DIALOG_INFO, getMessage("no.data.export"));
			return;
		}

		CSVDialog csvD = new CSVDialog(this, CSVDialog.EXPORT);
		csvD.pack();
		centreWindow(this, csvD);
		csvD.setVisible(true);
	}

	/** Import data from a CSV file */
	void importCSV()
	{
		CSVDialog csvD = new CSVDialog(this, CSVDialog.IMPORT);
		csvD.pack();
		centreWindow(this, csvD);
		csvD.show();
	}

	/** Export categories to a file */
	private void exportCategories()
	{
		Hashtable cats = def.getCategories();
		if (cats == null || cats.size() == 0)
		{
			InfoDialog.showInfoDialog(this, getMessage("error"), InfoDialog.DIALOG_INFO, getMessage("no.cats"));
			return;
		}

		CatExportDialog ced = new CatExportDialog(this);
		ced.pack();
		centreWindow(this, ced);
		ced.show();
	}

	/** Add or edit a card
	 * @param newCard set to true for a new card
	 * @param editing should be set to true if autoEdit is on or creating a new card
	 * @param copy should be set to true if using an existing card as a template
	 */
	public void editCard(boolean newCard, boolean editing, boolean copy)
	{
		CardDialog cDialog;

		// see if there are any fields defined
		if (def.getNumberOfFields() == 0)
		{
			InfoDialog.showInfoDialog(this, getMessage("error"), InfoDialog.DIALOG_INFO, getMessage("no.fields.def"));
			return;
		}

		if (newCard) {
			cDialog = new CardDialog(this, null, false, true, -1);
		} else {
			int selIndex = cardList.getSelectedIndex();
			if (selIndex < 0) return;
			Card card = cards.getCard(selIndex);
			cDialog = new CardDialog(this, card, copy, editing, selIndex);
		}
		cDialog.pack();
		// set to size of app if running on PDA, i.e. full screen
		if (settings.isPda()) {
			cDialog.setSize(getSize());
		}
		centreWindow(this, cDialog);
		cDialog.show();
	}

	/**
	 * While viewing a card the user wants to browse to the next/previous card
	 * @param direction either 1 or -1 for next or previous respectively
	 * @return the new selected index
	 */
	public int editOtherCard(int direction)
	{
		int selIndex = cardList.getSelectedIndex();
		if (selIndex < 0) return 0;
		selIndex += direction;
		int numRows = cardList.countItems();
		if (selIndex < 0) selIndex = numRows - 1;
		else if (selIndex >= numRows) selIndex = 0;
		cardList.select(selIndex);
		return selIndex;
	}

	/** Delete one or more cards */
	private void deleteCard()
	{
		int selection = cardList.getSelectedIndex();
		if (selection > -1)
		{
			cards.deleteCard(selection);
			if (selection > 0) {
				selection--;
			} else {
				selection = 0;
			}
			cardList.select(selection);
			changes = true;
		}
	}

	/** Restore the deleted cards */
	private void undeleteCard()
	{
		cards.undeleteCard();
		changes = true;
	}

	/** Display the category dialog window */
	public void editCategories()
	{
		CategoryDialog catDialog = new CategoryDialog(this, true);
		catDialog.pack();
		centreWindow(this, catDialog);
		catDialog.show();
	}

	/** Display the fields dialog window */
	private void editFields()
	{
		FieldsDialog fieldsDialog = new FieldsDialog(this);
		fieldsDialog.pack();
		centreWindow(this, fieldsDialog);
		fieldsDialog.show();
	}

	/**
	 * Get the minimum size of the app
	 * @return the minimum size
	 */
	public Dimension getMinimumSize()
	{
		return new Dimension(240, 282);
	}

	/**
	 * Get the preferred size of the app
	 * @return the preferred size
	 */
	public Dimension getPreferredSize()
	{
		int width, height;

		width = (settings.getOverrideWidth() == 0 ? 600 : settings.getOverrideWidth());
		height = (settings.getOverrideHeight() == 0 ? 400 : settings.getOverrideHeight());

		return new Dimension(width, height);
	}

	/** Get the maximum possible size */
	public Dimension getMaximumSize()
	{
		if (settings.isPda()) {
			return getPreferredSize();
		} else {
			Toolkit tk = Toolkit.getDefaultToolkit();
			return tk.getScreenSize();
		}
	}

	/** Exit the program */
	private void exit()
	{
		if (checkForChanges()) return;

		try	{
			saveProperties();
		} catch (Exception ex) {
			ex.printStackTrace();
		}

		setVisible(false);
		dispose();
		System.exit(0);
	}

	/**
	 * Centers any Window in the middle of the app or screen
	 * @param parent parent Window
	 * @param win the Window to centre
	 */
	public static void centreWindow(Window parent, Window win)
	{
		Dimension outer;
		if (!settings.isPda())
		{
			Point p = parent.getLocation();
			win.setLocation(p.x+50, p.y+50);
		}
		else
		{
			outer = parent.getSize();
			if (outer.height == 0 || outer.width == 0)
			{
				// main window hasn't been shown yet
				outer = new Dimension(settings.getOverrideWidth(), settings.getOverrideHeight());
			}
			Dimension window = win.getSize();
			int x = (outer.width - window.width) / 2;
			int y = (outer.height - window.height) / 2;
			win.setLocation(x, y);
		}
	}

	/**
	 * Edit user preferences
	 * @param visibleFields if true preferences dialog will default to visible fields tab
	 */
	void editPrefs(boolean visibleFields)
	{
		Dialog prefDialog = new PrefDialog(this, visibleFields);
		prefDialog.pack();
		centreWindow(this, prefDialog);
		prefDialog.show();

		cardList.setFont(baseFont);
		cardList.validate();
	}

	/** Save the properties to a file */
	private void saveProperties()
	{
		log("Saving properties", 9);
		Properties props = new Properties();

		// add preferences
		props.put(PREF_SAVE_CONF, (new Boolean(saveConfirmation)).toString());
		props.put(PREF_OPEN_LAST_FILE, (new Boolean(openLastFile)).toString());
		props.put(PREF_AUTO_EDIT, (new Boolean(autoEdit)).toString());
		props.put(PREF_WORD_WRAP, (new Boolean(wordWrap)).toString());
		String fontString = baseFont.getName()+","+baseFont.getStyle()+","+baseFont.getSize();
		props.put(PREF_FONT, fontString);
		props.put(PREF_TOOLBAR, (new Boolean(viewToolbar)).toString());

		// add last accessed files
		for (int i=0; i<lastFiles.length; i++) {
			if (lastFiles[i] != null) {
				props.put("lastFile"+i, lastFiles[i]);
			}
		}

		// save window position and size
		if (!settings.isPda())
		{
			Dimension size = getSize();
			Point location = getLocationOnScreen();
			props.put("size", size.width+"x"+size.height);
			props.put("location", location.x+","+location.y);
		}

		// save user icons
		Vector userIcons = settings.getUserIcons();
		if (userIcons.size() > 0)
		{
			I18NIcon icon;
			StringBuffer sb = new StringBuffer();
			Enumeration enm = userIcons.elements();
			while (enm.hasMoreElements())
			{
				icon = (I18NIcon)enm.nextElement();
				sb.append(icon.getKey());
				sb.append(",");
			}
			sb.setLength(sb.length()-1); // remove last ,
			props.put("icons", sb.toString());
		}

		try
		{
			File propsDir = new File(System.getProperty("user.home"), PROPS_DIR);
			propsDir.mkdirs();
			File saveFile = new File(propsDir, PROPS_FILE);
			FileOutputStream fos = new FileOutputStream(saveFile);
			props.save(fos, "JCards properties - do not edit");
		} catch (IOException ioe) {
			InfoDialog.showInfoDialog(this, getMessage("cannot.save"), InfoDialog.DIALOG_INFO, getMessage("props.save.failed"));
			log(ioe.getMessage(), 0);
		}
		log("Finished saving properties", 9);
	}

	/**
	 * Read any previously saved properties
	 * @return the last opened file or null if not specified or null if not required
	 */
	private String readProperties()
	{
		Properties props = new Properties();
		MenuItem lfMI;
		String lastFile = null;
		File f;
		int pos;

		log("Reading properties", 9);

		try
		{
			File home = new File(System.getProperty("user.home"));
			f = new File(home, PROPS_DIR+"/"+PROPS_FILE);
			log("Checking for properties file "+f.getAbsolutePath(), 7);
			if (!f.exists()) {
				log("No properties file", 6);
				return null;
			}

			log("Properties file exists", 5);
			FileInputStream fis = new FileInputStream(f);
			props.load(fis);

			// read preferences
			String savConf = props.getProperty(PREF_SAVE_CONF);
			if (savConf != null) {
				saveConfirmation = (new Boolean(savConf)).booleanValue();
			}
			log("Save confirmation="+saveConfirmation, 5);

			String opLast = props.getProperty(PREF_OPEN_LAST_FILE);
			if (opLast != null)	{
				openLastFile = (new Boolean(opLast)).booleanValue();
			}
			log("Open last file="+openLastFile, 5);

			String ae = props.getProperty(PREF_AUTO_EDIT);
			if (ae != null)	{
				autoEdit = (new Boolean(ae)).booleanValue();
			}
			log("Auto edit="+autoEdit, 5);

			String ww = props.getProperty(PREF_WORD_WRAP);
			if (ww != null) {
				wordWrap = (new Boolean(ww)).booleanValue();
			}
			log("Word wrap="+wordWrap, 5);

			String fontString = props.getProperty(PREF_FONT);
			if (fontString != null) {
				StringTokenizer st = new StringTokenizer(fontString, ",");
				if (st.countTokens() == 3)
				{
					try {
						baseFont = new Font(st.nextToken(), Integer.parseInt(st.nextToken()), Integer.parseInt(st.nextToken()));
					} catch (NumberFormatException nfex) {
						System.err.println("Bad font: "+fontString);
					}
				}
				cardList.setFont(baseFont);
				cardList.validate();
			}
			log("Font="+baseFont, 5);

			String tb = props.getProperty(PREF_TOOLBAR);
			if (tb != null) {
				viewToolbar = (new Boolean(tb)).booleanValue();
			}

			// update last opened menu
			log("Updating last opened menu", 9);
			for (int i=0; i<MAX_LAST_FILES; i++)
			{
				lastFile = props.getProperty("lastFile"+i);
				if (lastFile != null)
				{
					f = new File(lastFile);
					lfMI = new MenuItem(f.getName());
					lfMI.setActionCommand("LF:"+f.getAbsolutePath());
					lfMI.addActionListener(this);
					lastMenu.add(lfMI);
					lastFiles[i] = lastFile;
				}
			}
			log("Finished updating last opened menu", 9);

			// set size and position
			if (!settings.isPda())
			{
				log("Running on desktop", 5);
				String size = props.getProperty("size");
				if (size != null)
				{
					try
					{
						pos = size.indexOf("x");
						if (pos > 0)
						{
							String w = size.substring(0, pos);
							String h = size.substring(pos+1);
							settings.setOverrideWidth(Integer.parseInt(w));
							settings.setOverrideHeight(Integer.parseInt(h));
						}
					} catch (Exception ex) {}
				}
				String location = props.getProperty("location");
				if (location != null)
				{
					try
					{
						pos = location.indexOf(",");
						if (pos > 0)
						{
							String x = location.substring(0, pos);
							String y = location.substring(pos+1);
							settings.setInitLocationX(Integer.parseInt(x));
							settings.setInitLocationY(Integer.parseInt(y));
						}
					} catch (Exception ex) {}
				}
			}

			// read icons
			String icons = props.getProperty("icons");
			if (icons != null)
			{
				StringTokenizer st = new StringTokenizer(icons, ",");
				String[] iconArray = new String[st.countTokens()];
				int i = 0;
				while (st.hasMoreTokens()) {
					iconArray[i++] = st.nextToken();
				}
				settings.setUserIcons(iconArray);
				resetToolbar();
			}

			// load the last referenced file
			if (openLastFile) {
				lastFile = props.getProperty("lastFile"+0);
			} else {
				lastFile = null;
			}

		} catch (IOException ioe) {
			log("Error loading jca.properties: "+ioe.getMessage(), 0);
		}

		log("Finished reading properties", 9);

		return lastFile;
	}

	/**
	 * Add a card to the list
	 * @param card the new card to add
	 * @return the index of the new card
	 */
	public int addCard(Card card)
	{
		int index = cards.addCard(card);
		cardList.select(index);
		changes = true;

		return index;
	}

	/**
	 * Replace a card entry
	 * @param cardIndex the index of the card to replace
	 * @param card the replacement card
	 */
	public void replaceCard(int cardIndex, Card card)
	{
		int index = cards.setCard(cardIndex, card);
		cardList.select(index);
		changes = true;
	}

	/**
	 * Get the cards
	 */
	public Cards getCards()
	{
		return cards;
	}

	/** Edit the comment associated with the database */
	private void editComment()
	{
		CommentDialog commentDialog = new CommentDialog(this, cards);
		commentDialog.pack();
		centreWindow(this, commentDialog);
		commentDialog.show();
		commentDialog.dispose();
	}

	/**
	 * Search for matching cards
	 * @param search the search string
	 * @param field either 'All' or a particular field
	 * @param matchCase set to true for a case sensitive search
	 * @param searchInResults if true use the existing search results as the starting point
	 * @return true if some cards were found; false otherwise
	 */
	public boolean findCards(String search, String field, boolean matchCase, boolean searchInResults)
	{
		prevSearchText = search;
		prevSearchField = field;
		prevSearchMatchCase = matchCase;

		return cards.findCards(search, field, matchCase, searchInResults);
	}

	/** Load an image from either the JAR file or filesystem
	 * @param drawComp the component that the image will be drawn on
	 * @param imgName name of the image
	 * @return the loaded image
	 */
	public static Image getImage(Component drawComp, String imgName)
	{
		log("Loading image "+imgName, 5);
		Image img = null;
		URL imgURL = JCards.class.getResource(imgName);
		if (imgURL != null)
		{
			log("Loading image URL "+imgURL, 8);
			img = Toolkit.getDefaultToolkit().getImage(imgURL);
		}
		else
		{
			// remove leading /
			log("Loading image", 8);
			img = Toolkit.getDefaultToolkit().getImage(imgName.substring(1));
		}

		MediaTracker mt = new MediaTracker(drawComp);
		mt.addImage(img, 0);
		try	{
			mt.waitForID(0, 1000);
		} catch (InterruptedException ipex)	{
			log(ipex.getMessage(), 0);
		}

		return img;
	}

	/** Load an image from either the JAR file or filesystem
	 * @param imgName name of the image
	 * @return the loaded image
	 * @see #getImage(Component, String)
	 */
	public Image getImage(String imgName)
	{
		return getImage(this, imgName);
	}

	/** Show some information about the program */
	private void showAbout()
	{
		AboutDialog aboutDialog = new AboutDialog(this);
		aboutDialog.pack();
		centreWindow(this, aboutDialog);
		aboutDialog.show();
	}

	/** Show the password dialog and allow the user to enter/change a password */
	private void enterPassword()
	{
		PasswordDialog passDialog = new PasswordDialog(this, password, true);
		passDialog.pack();
		centreWindow(this, passDialog);
		passDialog.show();
	}

	/** Clear any password that is set on the database */
	private final void clearPassword()
	{
		password = "";
		changes = true;

		String[] lines = { getMessage("pass.cleared"), getMessage("file.unenc") };
		InfoDialog.showInfoDialog(this, getMessage("pass.cleared"), InfoDialog.DIALOG_INFO, lines);
	}

	/**
	 * Set the password used to encrypt the file
	 * @param newPassword new password
	 */
	public final void setPassword(String newPassword)
	{
		password = newPassword;
	}

	/**
	 * Get the password
	 * @param return the possibly null password
	 */
	public final String getPassword()
	{
		return password;
	}

	/**
	 * This method is called whenever the DBDef or Cards objects are changed
	 * JCards records the change
	 */
	public void update(Observable o, Object arg)
	{
		if (o instanceof DBDef)
		{
			String cmd = (String)arg;
			if (cmd.startsWith("Category") || cmd.startsWith("Field"))
			{
				updateViewMenu();
			}
		}
		else if (o instanceof Cards)
		{
			// don't register a find as a change
			CardUpdate cUpdate = (CardUpdate)arg;
			if (cUpdate.cardOperation == Cards.FIND) {
				return;
			}
		}

		changes = true;
	}

	/** Returns true if open last file is turned on
	 * @return true if open last file is turned on
	 * @see #setOpenLastFile(boolean)
	 */
	public boolean isOpenLastFile()
	{
		return openLastFile;
	}

	/** Set whether to open the last accessed file
	 * @param openLast set to true to open the last accessed file
	 * @see #isOpenLastFile()
	 */
	public void setOpenLastFile(boolean openLast)
	{
		openLastFile = openLast;
	}

	/** Returns true if save confirmation is turned on
	 * @return true if a dialog box will be displayed upon saving
	 * @see #setOpenLastFile(boolean)
	 */
	public boolean isSaveConfirmation()
	{
		return saveConfirmation;
	}

	/** Set whether a confirmation box should be displayed
	 * when saving
	 * @parm saveConf set to true to display a confirmation
	 * @see #isSaveConfirmation()
	 */
	public void setSaveConfirmation(boolean saveConf)
	{
		saveConfirmation = saveConf;
	}

	/** Returns true if cards are automatically edited upon opening
	 * @see #setAutoEdit(boolean)
	 */
	public boolean isAutoEdit()
	{
		return autoEdit;
	}

	/** Set whether cards should be automatically in edit mode when opened
	 * @param ae set to true to always edit cards
	 * @see #isAutoEdit()
	 */
	public void setAutoEdit(boolean ae)
	{
		autoEdit = ae;
	}

	/** Returns true if word wrap is on
	 ** @see #setWordWrap(boolean)
	 **/
	public boolean isWordWrap()
	{
		return wordWrap;
	}

	/** Set word wrap for memo fields
	 * @param ww set to true for word wrap
	 */
	public void setWordWrap(boolean ww)
	{
		wordWrap = ww;
	}

	/** Calculate the statistics for date and numeric columns */
	public void calculateStatistics()
	{
		FieldDef fdef;
		Vector statFields = new Vector();

		// check that there is some data
		if (cards.size() == 0)
		{
			InfoDialog.showInfoDialog(this, getMessage("no.data"), InfoDialog.DIALOG_INFO, getMessage("no.cards.def"));
			return;
		}

		// check that there are some date or numeric fields
		Vector fd = def.getFieldDefs();
		Enumeration enm = fd.elements();
		while (enm.hasMoreElements())
		{
			fdef = (FieldDef)enm.nextElement();
			switch (fdef.getType())
			{
				case FieldDef.DATE:
				case FieldDef.DOUBLE:
				case FieldDef.INTEGER:
					statFields.addElement(fdef.getName());
					break;
			}
		}

		if (statFields.size() == 0)
		{
			InfoDialog.showInfoDialog(this, getMessage("no.fields"), InfoDialog.DIALOG_INFO, getMessage("no.date.numeric"));
			return;
		}

		StatisticsDialog statDialog = new StatisticsDialog(this, statFields);
		statDialog.pack();
		centreWindow(this, statDialog);
		statDialog.show();
	}

	/**
	 * Show statistics for the selected field or for the database
	 * @param fieldName field selected by user
	 */
	public void showStatistics(String fieldName)
	{
		MessageFormat msgFormat;
		String stats[] = null;
		String title;

		if (fieldName != null)
		{
			FieldDef fd = def.getFieldDef(fieldName);
			int type = fd.getType();

			if (type == FieldDef.DATE) {
				stats = cards.getDateStatistics(fieldName);
			} else if (type == FieldDef.DOUBLE || type == FieldDef.INTEGER) {
				stats = cards.getNumericStatistics(fieldName);
			}

			// check for no data
			if (stats == null) {
				stats = new String[1];
				stats[0] = getMessage("no.data");
			}

			// set title
			msgFormat = new MessageFormat(getMessage("stats"));
			String[] args = { fieldName };
			title = msgFormat.format(args);
		}
		else
		{
			stats = new String[4];
			msgFormat = new MessageFormat(getMessage("db.recs"));
			Object[] args = {new Integer(cards.size())};
			stats[0] = msgFormat.format(args);
			msgFormat = new MessageFormat(getMessage("db.fields"));
			args[0] = new Integer(def.getNumberOfFields());
			stats[1] = msgFormat.format(args);
			msgFormat = new MessageFormat(getMessage("db.created"));
			args[0] = new Date(creationDate);
			stats[2] = msgFormat.format(args);
			msgFormat = new MessageFormat(getMessage("db.modified"));
			args[0] = new Date(modifiedDate);
			stats[3] = msgFormat.format(args);

			title = getMessage("stats.db");
		}

		InfoDialog.showInfoDialog(this, title, InfoDialog.DIALOG_INFO, stats);
	}

	/**
	 * Get the current file
	 * @param current file or null
	 */
	public File getCardFile() {
		return cardFile;
	}

	/**
	 * Edit the counters
	 */
	private void editCounters()
	{
		CounterDialog cDialog = new CounterDialog(this);
		cDialog.pack();
		centreWindow(this, cDialog);
		cDialog.show();
	}

	/** Get the current settings */
	public UserSettings getUserSettings()
	{
		return settings;
	}

	/** Set app to full screen */
	private void fullScreen()
	{
		Dimension setSize;
		Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
		if (restoreSize != null) {
			setSize = restoreSize;
			restoreSize = null;
			setSize(setSize);
			centreWindow(this, this);
		} else {
			restoreSize = getSize();
			setSize = screen;
			setLocation(0, 0);
			setSize(setSize);
		}
		settings.setOverrideHeight(setSize.height);
		settings.setOverrideWidth(setSize.width);
		invalidate();
		pack();
	}

	/** Copy a Card's contents to the clipboard */
	private void copyToClipboard()
	{
		int selIndex = cardList.getSelectedIndex();
		if (selIndex > -1)
		{
			Card c = cards.getCard(selIndex);
			Clipboard clip = Toolkit.getDefaultToolkit().getSystemClipboard();
			clip.setContents(c, null);
		}
	}

	/** Get the date the database was created */
	public long getCreationDate()
	{
		return creationDate;
	}

	/** Set the creation date */
	public void setCreationDate(long cDate)
	{
		creationDate = cDate;
	}

	/** Get the date the database was modified */
	public long getModifiedDate()
	{
		return modifiedDate;
	}

	/** Set the time when the database was last modified */
	public void setModifiedDate(long modDate)
	{
		modifiedDate = modDate;
	}

	/** Set the column widths */
	public void setColWidths(XMLElement colWidths)
	{
		cardList.setColWidths(colWidths);
	}

	/** Get the column widths */
	public XMLElement getColWidths()
	{
		return cardList.getColWidths();
	}

	/**
	 * Indicate that changes have been made to the database
	 */
	public void setChanges(boolean changes)
	{
		this.changes = changes;
	}

	/**
	 * Get an internationalised message
	 * @param key the key to the message
	 * @return the localised message
	 */
	public String getMessage(String key)
	{
		return msgs.getString(key);
	}

	/**
	 * Get the labels.properties ResourceBundle
	 * @return the labels ResourceBundle
	 */
	public static ResourceBundle getLabelsBundle() {
		return ResourceBundle.getBundle("nrs.jcards.props.labels");
	}

	/**
	 * Get the buttons.properties ResourceBundle
	 * @return the buttons ResourceBundle
	 */
	public static ResourceBundle getButtonsBundle() {
		return ResourceBundle.getBundle("nrs.jcards.props.buttons");
	}

	/**
	 * Get the menu.properties ResourceBundle
	 * @return the menu ResourceBundle
	 */
	public static ResourceBundle getMenuBundle() {
		return ResourceBundle.getBundle("nrs.jcards.props.menu");
	}

	/** User clicked close window decoration */
	public void windowClosing(WindowEvent e) {
		exit();
	}

	public void windowActivated(WindowEvent e) {}
	public void windowClosed(WindowEvent e) {}
	public void windowDeactivated(WindowEvent e) {}
	public void windowDeiconified(WindowEvent e) {}
	public void windowIconified(WindowEvent e) {}
	public void windowOpened(WindowEvent e) {}
}
