package nrs.jcards;

import java.util.Hashtable;
import java.util.Vector;
import java.util.ResourceBundle;

/*
 * This file is part of JCards.
 * JCards is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * JCards is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JCards; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * User settings
 *
 * @author Copyright &#169; 2002-2004 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class UserSettings
{
	public final static String[] DATE_FORMATS = {
		"dd/MM/yyyy", "dd/MM/yyyy HH:ss", "dd/MM/yy", "dd/MM/yy HH:ss",
		"MM/dd/yyyy", "MM/dd/yyyy HH:ss", "MM/dd/yy", "MM/dd/yy HH:ss",
		"dd.MM.yyyy", "dd.MM.yyyy HH:ss", "MM.yyyy",
		"yyyy", "dd MMM yyyy", "dd MMM yyyy HH:ss", "HH:ss",
		"yyyy-MM-dd", "yyyy-MM-dd HH:ss", "MM/dd", "MM/yy", "MM/yyyy"
	};
	
	/** Icons */
	public final static String ICON_ADD = "/pics/add.gif";
	public final static String ICON_CALCULATION = "/pics/calculation.gif"; // used in edit card dialog
	public final static String ICON_CANCEL = "/pics/cancel.gif";
	public final static String ICON_CATEGORIES = "/pics/categories.gif";
	public final static String ICON_CLIPBOARD = "/pics/clipboard.gif";
	public final static String ICON_COPY = "/pics/copy.gif";
	public final static String ICON_COUNTER = "/pics/counter.gif";
	public final static String ICON_DELETE = "/pics/remove.gif";
	public final static String ICON_DOWN = "/pics/down.gif";
	public final static String ICON_EDIT = "/pics/edit.gif";
	public final static String ICON_EXIT = "/pics/exit.gif";
	public final static String ICON_FIELDS = "/pics/fields.gif";
	public final static String ICON_FIND = "/pics/find.gif";
	public final static String ICON_FULLSCREEN = "/pics/fullscreen.gif";
	public final static String ICON_NEW = "/pics/new.gif";
	public final static String ICON_NEXT = "/pics/next.gif";
	public final static String ICON_OK = "/pics/ok.gif";
	public final static String ICON_OPEN = "/pics/open.gif";
	public final static String ICON_PASSWORD = "/pics/password.gif";
	public final static String ICON_PREFERENCES = "/pics/configure.gif";
	public final static String ICON_PREVIOUS = "/pics/previous.gif";
	public final static String ICON_RESET = "/pics/reload.gif";
	public final static String ICON_SAVE = "/pics/save.gif";
	public final static String ICON_SAVE_AS = "/pics/saveas.gif";
	public final static String ICON_UNDELETE = "/pics/undelete.gif";
	public final static String ICON_UP = "/pics/up.gif";
	
	/** Environment settings */
	public boolean pda = false;
	public int overrideWidth = 0;
	public int overrideHeight = 0;
	private int initLocationX = -1;
	private int initLocationY = -1;
	private Vector userIcons = null;
	
	/** Creates a new instance of UserSettings */
	public UserSettings()
	{
		// setup initial icons
		resetIcons();
	}
	
	/** Set default icons appropriate to where the program is running */
	public void resetIcons()
	{
		// these are not defined as final static class variables to save memory
		if (pda)
		{
			String[] pdaIcons = { "save", "card.new", "card.edit", "card.del", "fields", "categories" };
			setUserIcons(pdaIcons);
		}
		else
		{
			String[] desktopIcons = {
				"open", "save", "card.new", "card.edit",
				"card.del", "find", "fields", "categories", "prefs"
			};			
			setUserIcons(desktopIcons);
		}
	}
	
	/**
	 * Setup a hashmap of names to icons for all icons
	 * @param keyOnName if true the Hashtable key is the localised name, if false the key
	 * @return a Hashtable of icon mappings
	 * @see I18NIcon
	 */
	public Hashtable getAllIcons(boolean keyOnName)
	{
		// not defined statically to minimise memory
		String[][] defs = {
			{ "new", ICON_NEW },
			{ "exit", ICON_EXIT },
			{ "card.del", ICON_DELETE },
			{ "card.edit", ICON_EDIT },
			{ "card.new", ICON_ADD },
			{ "card.undel", ICON_UNDELETE },
			{ "card.copy", ICON_COPY },
			{ "copy.clip", ICON_CLIPBOARD },
			{ "categories", ICON_CATEGORIES },
			{ "counters", ICON_COUNTER },
			{ "fields", ICON_FIELDS },
			{ "find", ICON_FIND },
			{ "full.screen", ICON_FULLSCREEN },
			{ "open", ICON_OPEN },
			{ "prefs", ICON_PREFERENCES },
			{ "save", ICON_SAVE },
			{ "save.as", ICON_SAVE_AS },
			{ "pass.set", ICON_PASSWORD }
		};
		
		// convert to I18NIcons
		// icons equate to menu items which is why they are not redefined in the buttons bundle
		ResourceBundle iconNames = JCards.getMenuBundle();
		Hashtable icons = new Hashtable(defs.length);
		I18NIcon icon;
		for (int i=0; i<defs.length; i++)
		{
			icon = new I18NIcon();
			icon.setKey(defs[i][0]);
			icon.setName(iconNames.getString(defs[i][0]));
			icon.setGraphic(defs[i][1]);
			if (keyOnName) {
				icons.put(icon.getName(), icon);
			} else {
				icons.put(icon.getKey(), icon);
			}
		}
		
		return icons;
	}

	/**
	 * Set the user icons
	 * @param keys array of icon key names
	 */
	public void setUserIcons(String[] keys)
	{
		if (keys.length == 0) return;
		
		I18NIcon icon;
		String key;
		userIcons = new Vector(keys.length);
		Hashtable allIcons = getAllIcons(false);
		
		for (int i=0; i<keys.length; i++)
		{
			key = keys[i];
			icon = (I18NIcon)allIcons.get(key);
			if (icon != null) {
				userIcons.addElement(icon);
			}
		}
	}
	
	/** Return true if running on a PDA */
	public boolean isPda() {
		return pda;
	}
	
	/** Set whether running on a PDA or desktop */
	public void setPda(boolean pda) {
		this.pda = pda;
		resetIcons();
	}
	
	public void setOverrideWidth(int overrideWidth) {
		this.overrideWidth = overrideWidth;
	}
	
	public void setOverrideHeight(int overrideHeight) {
		this.overrideHeight = overrideHeight;
	}
	
	public int getInitLocationX() {
		return initLocationX;
	}
	
	public void setInitLocationX(int initLocationX) {
		this.initLocationX = initLocationX;
	}
	
	public int getInitLocationY() {
		return initLocationY;
	}
	
	public void setInitLocationY(int initLocationY) {
		this.initLocationY = initLocationY;
	}
	
	/** Get the icons currently selected by the user */
	public Vector getUserIcons() {
		return userIcons;
	}
	
	/**
	 * Set the icons that the user wants to see
	 * @param newUserIcons array of localised menu/icon names
	 */
	public void setUserIconNames(String[] newUserIcons)
	{
		if (newUserIcons.length == 0) return;
		
		Hashtable icons = getAllIcons(true);
		userIcons = new Vector(newUserIcons.length);
		String name;
		I18NIcon icon;
		
		for (int i=0; i<newUserIcons.length; i++)
		{
			name = newUserIcons[i];
			icon = (I18NIcon)icons.get(name);
			if (icon != null) {
				userIcons.addElement(icon);
			}
		}
	}
	
	public int getOverrideWidth() {
		return overrideWidth;
	}
	
	public int getOverrideHeight() {
		return overrideHeight;
	}
}
