package nrs.jcards.dialog;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.File;
import java.util.ResourceBundle;
import jclass.bwt.BWTEnum;
import jclass.bwt.JCActionEvent;
import jclass.bwt.JCActionListener;
import jclass.bwt.JCButton;
import nrs.jcards.DBDef;
import nrs.jcards.JCards;
import nrs.jcards.UserSettings;
import nrs.jcards.io.CSVIO;

/*
 * JCards
 * Copyright (C) 2004 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * Dialog for importing and export CSV files
 *
 * @author Copyright &#169; 2002-2004 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class CSVDialog extends Dialog implements ActionListener, JCActionListener, WindowListener
{
	/** Used in the constructor to specify the dialog type */
	public final static int IMPORT = 0;
	public final static int EXPORT = 1;

	/** Field delimiters */
	final static String[] FIELD_DELIM = { ",", "^", "$", "!", "#", ";", ":", "?", "/", "*", "|", "<space>" };
	/** Text delimiters */
	private final static String[] TEXT_DELIM = { "\"", "^", "$", ":", "?", "/", "*", "|", "<none>" };

	/** Character sets */
	final static String[] CHAR_SETS = { "ASCII", "UTF8", "ISO8859_1",
		"ISO8859_2", "ISO8859_3", "ISO8859_4", "ISO8859_5", "ISO8859_6",
		"ISO8859_7", "ISO8859_8", "ISO8859_9", "KOI8_R", "JIS0201", "JIS0208",
		"JIS0212", "SJIS", "ISO2022KR" };

	/** Character sequence used to represent a newline */
	public final static String NEWLINE = "XX";

	private JCards jcards = null;
	private DBDef def = DBDef.getInstance();
	private int type = IMPORT;
	private Checkbox headerRow = null;
	private Checkbox concatNL = null;
	private Choice fieldDelimiter, textDelimiter, charSet;
	private TextField csvFile;

	/**
	 * Creates new form ImportDialog
	 * @param parent frame
	 * @param type import or export
	 */
	public CSVDialog(JCards parent, int type)
	{
		super((Frame)parent, true);

		int i;
		ResourceBundle buttons = JCards.getButtonsBundle();
		ResourceBundle labels = JCards.getLabelsBundle();

		jcards = parent;
		this.type = type;
		if (type == IMPORT) setTitle(labels.getString("csv.import"));
		else setTitle(labels.getString("csv.export"));

		setFont(jcards.baseFont);
		
		addWindowListener(this);

		Panel selectionPanel = new Panel(new GridBagLayout());
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.gridwidth = 1;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.insets = new Insets(5, 5, 5, 5);
		
        Label label = new Label(labels.getString("field.delimiter"));
        selectionPanel.add(label, gbc);

		fieldDelimiter = new Choice();
		for (i=0; i<FIELD_DELIM.length; i++) {
			fieldDelimiter.add(FIELD_DELIM[i]);
		}
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        selectionPanel.add(fieldDelimiter, gbc);

		label = new Label(labels.getString("text.delimiter"));
        gbc.gridwidth = 1;
        selectionPanel.add(label, gbc);

		textDelimiter = new Choice();
		for (i=0; i<TEXT_DELIM.length; i++) {
			textDelimiter.add(TEXT_DELIM[i]);
		}
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        selectionPanel.add(textDelimiter, gbc);

		label = new Label(labels.getString("file"));
		gbc.gridwidth = 1;
        selectionPanel.add(label, gbc);

		csvFile = new TextField(10);
        gbc.fill = GridBagConstraints.HORIZONTAL;
        gbc.weightx = 1.0;
        selectionPanel.add(csvFile, gbc);

		Button browseButton = new Button(buttons.getString("browse"));
		browseButton.setActionCommand("browse");
		browseButton.addActionListener(this);
        gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.weightx = 0.0;
		gbc.fill = GridBagConstraints.NONE;
        selectionPanel.add(browseButton, gbc);

		label = new Label(labels.getString("charset"));
		gbc.gridwidth = 1;
        selectionPanel.add(label, gbc);

		charSet = new Choice();
		for (i=0; i<CHAR_SETS.length; i++) {
			charSet.add(CHAR_SETS[i]);
		}
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        selectionPanel.add(charSet, gbc);

		gbc.gridwidth = 1;

		Label headerLabel = new Label(labels.getString("header.row"));
		selectionPanel.add(headerLabel, gbc);

		headerRow = new Checkbox();
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		selectionPanel.add(headerRow, gbc);

		if (type == EXPORT)
		{
			Label nl = new Label(labels.getString("remove.newlines"));
			gbc.gridwidth = 1;
			selectionPanel.add(nl, gbc);

			concatNL = new Checkbox();
			gbc.gridwidth = GridBagConstraints.REMAINDER;
			selectionPanel.add(concatNL, gbc);

			if (jcards.getCardFile() != null)
			{
				String fileName = jcards.getCardFile().getAbsolutePath();
				int dotIndex = fileName.lastIndexOf(".");
				String expName;
				if (dotIndex > -1) {
					expName = fileName.substring(0, dotIndex)+".csv";
				} else {
					expName = fileName + ".csv";
				}
				csvFile.setText(expName);
			}
		}
        add(selectionPanel, BorderLayout.CENTER);

		// buttons
		Panel buttonPanel = new Panel(new FlowLayout(FlowLayout.RIGHT, 5, 5));

		JCButton cancelButton = new JCButton(buttons.getString("cancel"), JCards.getImage(this, UserSettings.ICON_CANCEL), BWTEnum.STRING_RIGHT);
        cancelButton.setActionCommand("cancel");
		cancelButton.addActionListener(this);
        buttonPanel.add(cancelButton);

		JCButton okButton = new JCButton(buttons.getString("ok"), JCards.getImage(this, UserSettings.ICON_OK), BWTEnum.STRING_RIGHT);
        okButton.setActionCommand("ok");
		okButton.addActionListener(this);
        buttonPanel.add(okButton);

        add(buttonPanel, BorderLayout.SOUTH);
	}

	public void pack()
	{
		super.pack();
		csvFile.setCaretPosition(csvFile.getText().length());
		csvFile.requestFocus();
	}

	/** Invoked when an action occurs. */
	public void actionPerformed(ActionEvent e)
	{
		String ac = e.getActionCommand();

		if (ac.equals("browse")) selectFile();
		else if (ac.equals("cancel")) windowClosing(null);
		else if (ac.equals("ok")) processCSV();
	}

	/** User clicked OK or cancel */
	public void actionPerformed(JCActionEvent e)
	{
		actionPerformed(new ActionEvent(e.getSource(), e.getID(), e.getActionCommand()));
	}

	/** Select a file */
	private void selectFile()
	{
		int mode = (type == IMPORT ? FileDialog.LOAD : FileDialog.SAVE);
		FileDialog fDialog = new FileDialog(jcards, jcards.getMessage("csv.select"), mode);
		fDialog.show();
		String file = fDialog.getFile();
		String dir = fDialog.getDirectory();
		fDialog.dispose();

		if (file != null)
		{
			file = dir + file;
			csvFile.setText(file);
			csvFile.setCaretPosition(file.length());
		}
	}
	
	/** Validate the user's actions */
	private void processCSV()
	{
		String td = textDelimiter.getSelectedItem();
		String fd = fieldDelimiter.getSelectedItem();

		if (td.equals(fd)) {
			InfoDialog.showInfoDialog(jcards, jcards.getMessage("error"), InfoDialog.DIALOG_INFO, jcards.getMessage("same.field.text.delim"));
			return;
		}

		JCards.log("Field delimiter="+fd, 7);
		JCards.log("Character delimiter="+td, 7);

		String file = csvFile.getText();
		if (file.length() > 0)
		{
			File csv = new File(file);

			boolean hasHeader = headerRow.getState();
			String cs = charSet.getSelectedItem();
			fd = fd.substring(0, 1);
			td = td.substring(0, 1);
			if (fd.equals("<")) {
				fd = " ";
			}
			if (td.equals("<")) {
				td = "";
			}

			if (type == EXPORT) {
				exportCSV(csv, fd, td, hasHeader, cs);
			} else if (type == IMPORT) {
				importCSV(csv, fd, td, hasHeader, cs);
			}
		}
		else
		{
			InfoDialog.showInfoDialog(jcards, jcards.getMessage("error"), InfoDialog.DIALOG_INFO, jcards.getMessage("no.file"));
		}
	}

	/**
	 * Import a CSV file
	 * @param file the file to import
	 * @param fDelim file delimiter
	 * @param tDelim text delimiter
	 * @param hasHeader CSV file has field names as first row
	 * @param charset character set
	 */
	private void importCSV(File file, String fDelim, String tDelim, boolean hasHeader, String charset)
	{
		JCards.log("Importing CSV", 9);

		// check file exists
		if (!file.exists())
		{
			InfoDialog.showInfoDialog(jcards, jcards.getMessage("error"), InfoDialog.DIALOG_INFO, jcards.getMessage("no.file"));
			return;
		}

		CSVIO io = new CSVIO(jcards);
		io.importCSV(file, fDelim, tDelim, charset, hasHeader);
		
		JCards.log("Finished importing CSV", 9);
		
		// close the dialog
		windowClosing(null);
	}
	
	/**
	 * Export data to a CSV file
	 * @param file the file to export to
	 * @param fDelim file delimiter
	 * @param tDelim text delimiter
	 * @param hasHeader CSV file has field names as first row
	 * @param charset character set
	 */
	private void exportCSV(File file, String fDelim, String tDelim, boolean hasHeader, String charset)
	{
		JCards.log("Exporting CSV", 9);

		if (file.exists())
		{
			int ret = InfoDialog.showInfoDialog(jcards, jcards.getMessage("confirm"), InfoDialog.DIALOG_CONFIRM, jcards.getMessage("file.exists"));
			if (ret != InfoDialog.YES) return;
		}

		boolean concat = concatNL.getState();
		
		CSVIO io = new CSVIO(jcards);
		io.exportCSV(file, fDelim, tDelim, charset, concat, hasHeader);
		
		JCards.log("Finished exporting CSV", 9);
		
		windowClosing(null);
	}
	
	/** Remove the dialog window */
	public void windowClosing(WindowEvent e)
	{
		dispose();
	}

	public void windowActivated(WindowEvent e) {}
	public void windowClosed(WindowEvent e) {}
	public void windowDeactivated(WindowEvent e) {}
	public void windowDeiconified(WindowEvent e) {}
	public void windowIconified(WindowEvent e) {}
	public void windowOpened(WindowEvent e) {}	
}
