package nrs.jcards.dialog;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.text.MessageFormat;
import java.text.ParseException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.ResourceBundle;
import java.util.Vector;

import nrs.jcards.Card;
import nrs.jcards.DBDef;
import nrs.jcards.FieldDef;
import nrs.jcards.InvalidFieldException;
import nrs.jcards.JCards;
import nrs.jcards.DisplayCardPanel;
import nrs.jcards.UserSettings;
import nrs.jcards.dialog.InfoDialog;

import jclass.bwt.BWTEnum;
import jclass.bwt.JCActionEvent;
import jclass.bwt.JCActionListener;
import jclass.bwt.JCButton;


/*
 * This file is part of JCards.
 * JCards is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * JCards is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JCards; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * Card Dialog allows the entry of new cards and the editing of existing cards
 *
 * @author Copyright &#169; 2002-2004 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class CardDialog extends Dialog implements ActionListener, JCActionListener, WindowListener
{
	private JCards jcards;
	private int cardIndex;
	private boolean editing = false;
	private boolean copy = false;
	private DisplayCardPanel dcp = null;

	/**
	 * Creates a new instance of CardDialog
	 * @param jcards the parent frame
	 * @param card Card that is being edited, may be null for a new one
	 * @param copy set to true if copying a card
	 * @param editing set to true to edit card
	 * @param cardSelection if editing this is set to the card currently
	 *			    selected, otherwise -1 for a new card
	 */
	public CardDialog(JCards jcards, Card card, boolean copy, boolean editing, int cardSelection)
	{
		super(jcards, true);

		ResourceBundle buttons = JCards.getButtonsBundle();
		ResourceBundle labels = JCards.getLabelsBundle();

		setTitle(labels.getString("edit.card"));

		this.jcards = jcards;
		this.copy = copy;
		this.editing = editing;
		cardIndex = cardSelection;

		setLayout(new BorderLayout());

		boolean newCard = (card == null ? true : false);
		dcp = new DisplayCardPanel(jcards, newCard);
		dcp.display(card, editing, jcards.isWordWrap(), this);
		ScrollPane scrollPane = new ScrollPane();
		scrollPane.add(dcp);
		add(scrollPane, BorderLayout.CENTER);

		// add toolbar
		Panel toolbar = new Panel(new FlowLayout(FlowLayout.LEFT, 1, 0));
		add(toolbar, BorderLayout.NORTH);

		JCButton button;
		if (editing)
		{
			JCButton ok = new JCButton(buttons.getString("ok"), jcards.getImage(UserSettings.ICON_OK), BWTEnum.STRING_RIGHT);
			ok.setActionCommand("OK");
			ok.addActionListener(this);
			toolbar.add(ok);

			// only need an update button if calculated fields are defined
			if (DBDef.getInstance().areCalculationsDefined())
			{
				JCButton update = new JCButton(buttons.getString("update"), jcards.getImage(UserSettings.ICON_CALCULATION), BWTEnum.STRING_RIGHT);
				update.setActionCommand("update");
				update.addActionListener(this);
				toolbar.add(update);
			}
		}
		else
		{
			JCButton edit = new JCButton(buttons.getString("edit"), jcards.getImage(UserSettings.ICON_EDIT), BWTEnum.STRING_RIGHT);
			edit.setActionCommand("edit");
			edit.addActionListener(this);
			toolbar.add(edit);
		}

		JCButton cancel = new JCButton(buttons.getString("cancel"), jcards.getImage(UserSettings.ICON_CANCEL), BWTEnum.STRING_RIGHT);
		cancel.setActionCommand("cancel");
		cancel.addActionListener(this);
		toolbar.add(cancel);

		if (!editing)
		{
			JCButton previous = new JCButton(null, jcards.getImage(UserSettings.ICON_PREVIOUS), BWTEnum.STRING_LEFT);
			previous.setActionCommand("previous");
			previous.addActionListener(this);
			toolbar.add(previous);

			JCButton next = new JCButton(null, jcards.getImage(UserSettings.ICON_NEXT), BWTEnum.STRING_LEFT);
			next.setActionCommand("next");
			next.addActionListener(this);
			toolbar.add(next);
		}

		// when user clicks close window decoration
		addWindowListener(this);
	}

	/** @see #actionPerformed(ActionEvent e) */
	public void actionPerformed(JCActionEvent e)
	{
		ActionEvent ae = new ActionEvent(e.getSource(), e.getID(), e.getActionCommand());
		actionPerformed(ae);
	}

	/** Invoked when one of the buttons is clicked */
	public void actionPerformed(ActionEvent e)
	{
		String cmd = e.getActionCommand();

		// category edit command
		if (cmd.startsWith("..."))
		{
			jcards.editCategories();

			String field = cmd.substring(3);
			Choice ch = (Choice)dcp.getEntries().get(field);
			ch.setVisible(false);
			// we can't do removeAll as it blows up under Motif
			int count = ch.getItemCount();
			for (int i=count; i>0; i--)
			{
				ch.remove(i-1);
			}
			// add blank choice again followed by other choices
			ch.add("");
			FieldDef fd = DBDef.getInstance().getFieldDef(field);
			Vector catVec = DBDef.getInstance().getCategory(fd.getName());
			if (catVec != null)
			{
				Enumeration enm = catVec.elements();
				while (enm.hasMoreElements())
				{
					ch.add((String)enm.nextElement());
				}
				ch.setVisible(true);
			}
			return;
		}

		// CardDialog command
		if (cmd.equals("OK"))
		{
			if (DBDef.getInstance().areCalculationsDefined())
			{
				try {
					dcp.updateCalculations();
				} catch (InvalidFieldException ifex) {
					String[] msgLines = new String[2];
					msgLines[0] = ifex.getMessage();
					msgLines[1] = jcards.getMessage("continue.save");
					if (InfoDialog.showInfoDialog(jcards, jcards.getMessage("calc.error"), InfoDialog.DIALOG_CONFIRM, msgLines) != InfoDialog.YES) {
						return;
					}
				}
			}

			if (!updateValues()) {
				return;
			}
			setVisible(false);
			dispose();
		}
		else if (cmd.equals("edit"))
		{
			setVisible(false);
			dispose();
			jcards.editCard(false, true, false);
		}
		else if (cmd.equals("next"))
		{
			int cardSelection = jcards.editOtherCard(1);
			Card card = jcards.getCards().getCard(cardSelection);
			dcp.display(card, editing, jcards.isWordWrap(), this);
		}
		else if (cmd.equals("previous"))
		{
			int cardSelection = jcards.editOtherCard(-1);
			Card card = jcards.getCards().getCard(cardSelection);
			dcp.display(card, editing, jcards.isWordWrap(), this);
		}
		else if (cmd.equals("update"))
		{
			try {
				dcp.updateCalculations();
			} catch (InvalidFieldException ifex) {
				InfoDialog.showInfoDialog(jcards, jcards.getMessage("error"), InfoDialog.DIALOG_INFO, ifex.getMessage());
			}
		}
		else if (cmd.equals("cancel"))
		{
			if (editing) {
				dcp.rollbackCounters();
			}
			setVisible(false);
			dispose();
		}
	}

	/**
	 * Update the values
	 * @return true if updated successfully
	 */
	private boolean updateValues()
	{
		Hashtable entries = dcp.getEntries();
		Hashtable values = new Hashtable(entries.size());
		Object obj;
		Checkbox cb;
		Choice ch;
		TextComponent tc;
		Component comp;
		FieldDef fd;
		String tex, fieldName;
		DBDef def = DBDef.getInstance();
		String dateFormat = def.getDateFormat();
		int fieldType;

		Enumeration enm = entries.keys();
		while (enm.hasMoreElements())
		{
			fieldName = (String)enm.nextElement();
			fd = DBDef.getInstance().getFieldDef(fieldName);
			fieldType = fd.getType();
			comp = (Component)entries.get(fieldName);

			// check for mandatory entries
			// boolean, counter and calculated fields can't be mandatory
			if (fieldType != FieldDef.BOOLEAN && fieldType != FieldDef.COUNTER && fieldType != FieldDef.CALCULATED)
			{
				if (fieldType != FieldDef.CATEGORY)
				{
					tc = (TextComponent)comp;

					if (fd.isMandatory() && tc.getText().length() == 0)
					{
						MessageFormat msgFormat = new MessageFormat(jcards.getMessage("mandatory.field.msg"));
						String[] args = { fd.getName() };
						InfoDialog.showInfoDialog(jcards, jcards.getMessage("mandatory.field"), InfoDialog.DIALOG_INFO, msgFormat.format(args));
						comp.requestFocus();
						return false;
					}

					// only update values if something has been entered
					if (tc.getText().length() == 0)	{
						continue;
					}

					// ignore blank dates as well
					if (fieldType == FieldDef.DATE && tc.getText().equals(dateFormat)) {
						continue;
					}
				}
				else
				{
					ch = (Choice)comp;

					if (fd.isMandatory() && ch.getSelectedIndex() == 0)
					{
						MessageFormat msgFormat = new MessageFormat(jcards.getMessage("mandatory.field.msg"));
						String[] args = { fd.getName() };
						InfoDialog.showInfoDialog(jcards, jcards.getMessage("mandatory.field"), InfoDialog.DIALOG_INFO, msgFormat.format(args));
						comp.requestFocus();
						return false;
					}
				}
			}

			switch (fieldType)
			{
				case FieldDef.BOOLEAN:
					cb = (Checkbox)comp;
					values.put(fd.getName(), new Boolean(cb.getState()));
					break;

				case FieldDef.STRING:
				case FieldDef.MEMO:
					tc = (TextComponent)comp;
					values.put(fd.getName(), tc.getText());
					break;

				case FieldDef.INTEGER:
					tc = (TextComponent)comp;
					try	{
						values.put(fd.getName(), Integer.valueOf(tc.getText()));
					} catch (NumberFormatException nfex) {
						MessageFormat msgFormat = new MessageFormat(jcards.getMessage("incorrect.value.int"));
						String[] args = { tc.getText() };
						InfoDialog.showInfoDialog(jcards, jcards.getMessage("incorrect.value"), InfoDialog.DIALOG_INFO, msgFormat.format(args));
						tc.requestFocus();
						return false;
					}
					break;

				case FieldDef.DOUBLE:
					tc = (TextComponent)comp;
					try	{
						values.put(fd.getName(), Double.valueOf(tc.getText()));
					} catch (NumberFormatException nfex) {
						MessageFormat msgFormat = new MessageFormat(jcards.getMessage("incorrect.value.double"));
						String[] args = { tc.getText() };
						InfoDialog.showInfoDialog(jcards, jcards.getMessage("incorrect.value"), InfoDialog.DIALOG_INFO, msgFormat.format(args));
						tc.requestFocus();
						return false;
					}
					break;

				case FieldDef.DATE:
					tc = (TextComponent)comp;
					try	{
						values.put(fd.getName(), def.parseDate(tc.getText()));
					} catch (ParseException pex) {
						MessageFormat msgFormat = new MessageFormat(jcards.getMessage("incorrect.value.date"));
						Object[] args = { tc.getText(), dateFormat };
						 InfoDialog.showInfoDialog(jcards, jcards.getMessage("incorrect.value"), InfoDialog.DIALOG_INFO, msgFormat.format(args));
						 tc.requestFocus();
						 return false;
					}
					break;

				case FieldDef.CATEGORY:
					ch = (Choice)comp;
					values.put(fd.getName(), ch.getSelectedItem());
					break;

				case FieldDef.COUNTER:
				case FieldDef.CALCULATED:
					Label l = (Label)comp;
					if (l.getText().length() > 0) {
						if (l.getText().equals(DisplayCardPanel.ERROR)) {
							values.put(fd.getName(), new Double(DisplayCardPanel.ERROR_NUM));
						} else {
							values.put(fd.getName(), new Double(l.getText()));
						}
					}
					break;

				default:
					System.err.println("Unknown field type:"+fd.getType());
					return false;
			}
		}

		if (copy || cardIndex < 0) {
			jcards.addCard(new Card(values));
		} else {
			jcards.replaceCard(cardIndex, new Card(values));
		}

		return true;
	}

	/**
	 * Get the preferred size of the CardDialog
	 * @return Dimension
	 */
	public Dimension getPreferredSize()
	{
		if (jcards.getUserSettings().isPda()) {
			return jcards.getSize();
		} else {
			return new Dimension(500, 400);
		}
	}

	/** User clicked the close window decoration */
	public void windowClosing(WindowEvent e) {
		setVisible(false);
		dispose();
	}

	public void windowActivated(WindowEvent e) {}
	public void windowClosed(WindowEvent e) {}
	public void windowDeactivated(WindowEvent e) {}
	public void windowDeiconified(WindowEvent e) {}
	public void windowIconified(WindowEvent e) {}
	public void windowOpened(WindowEvent e) {
		dcp.setMemoStart();
	}
}
