package nrs.jcards.dialog;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.ResourceBundle;
import java.util.StringTokenizer;

import nrs.jcards.JCards;

/*
 * JCards
 * Copyright (C) 2004 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * Presents an information dialog to the user 
 *
 * @author Copyright &#169; 2002-2004 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class InfoDialog extends Dialog implements ActionListener, WindowListener
{
	private int response = 0;
	
	/** Types of dialog box */
	public final static int DIALOG_INFO = 0;
	public final static int DIALOG_CONFIRM = 1;
	
	/** Responses from dialog box */
	public final static int YES = 0;
	public final static int NO = 1;
	public final static int CANCEL = 2;

	/** Newline message character */
	private final static String NEW_LINE_CHAR = "*";
	
	/** Creates a new instance of InfoDialog for a Frame parent */
	private InfoDialog(Frame parentFrame, String title) {
		super(parentFrame, title, true);
		
		setFont(JCards.baseFont);
	}

	/**
	 * Show an info dialog, similar to JOptionPane
	 * @param parentFrame the owning Frame
	 * @param title window title
	 * @param dialogType either DIALOG_INFO or DIALOG_CONFIRM
	 * @param messageLine a line to display. If this contains the newline character (*) it will be split into multiple lines
	 * @return one of YES, NO or CANCEL
	 * @see #showInfoDialog(Frame, String, int, String[])
	 */
	public static int showInfoDialog(Frame parentFrame, String title, int dialogType, String messageLine)
	{
		StringTokenizer st = new StringTokenizer(messageLine, NEW_LINE_CHAR);
		int tokens = st.countTokens();
		String[] msgLines;
		if (tokens > 0) {
			int i = 0;
			msgLines = new String[tokens];
			while (st.hasMoreTokens()) {
				msgLines[i++] = st.nextToken();
			}
		} else {
			msgLines = new String[1];
			msgLines[0] = messageLine;
		}

		return showInfoDialog(parentFrame, title, dialogType, msgLines);
	}
	
	/**
	 * Show an info dialog, similar to JOptionPane
	 * @param parentFrame the owning Frame
	 * @param title window title
	 * @param dialogType either DIALOG_INFO or DIALOG_CONFIRM
	 * @param messageLines each element of the array is displayed on a separate line
	 * @return one of YES, NO or CANCEL
	 */
	public static int showInfoDialog(Frame parentFrame, String title, int dialogType, String[] messageLines)
	{
		InfoDialog infoDialog = new InfoDialog(parentFrame, title);
		
		ResourceBundle buttonLabels = JCards.getButtonsBundle();
		Label l;
		GridBagLayout gbl = new GridBagLayout();
		GridBagConstraints gbc = new GridBagConstraints();
		
		infoDialog.setLayout(gbl);
		
		gbc.insets = new Insets(2, 2, 2, 2);
		gbc.anchor = GridBagConstraints.CENTER;
		gbc.fill = GridBagConstraints.BOTH;
		gbc.weightx = 1.0;
		gbc.weighty = 1.0;
		if (dialogType == DIALOG_CONFIRM) {
			gbc.gridwidth = 3;
		} else {
			gbc.gridwidth = GridBagConstraints.REMAINDER;
		}
		
		for (int i=0; i<messageLines.length; i++)
		{
			l = new Label(messageLines[i], Label.CENTER);
			gbc.gridy = i;
			gbl.setConstraints(l, gbc);
			infoDialog.add(l);
		}
		
		gbc.gridy = messageLines.length;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbc.anchor = GridBagConstraints.SOUTH;
		switch (dialogType)
		{
			case DIALOG_CONFIRM:
				Button yesButton = new Button(buttonLabels.getString("yes"));
				yesButton.setActionCommand("DIALOG_YES");
				yesButton.addActionListener(infoDialog);
				gbc.gridwidth = 1;
				gbc.weighty = 0.0;
				gbl.setConstraints(yesButton, gbc);
				infoDialog.add(yesButton);
				
				Button noButton = new Button(buttonLabels.getString("no"));
				noButton.setActionCommand("DIALOG_NO");
				noButton.addActionListener(infoDialog);
				gbl.setConstraints(noButton, gbc);
				infoDialog.add(noButton);
				
				Button cancelButton = new Button(buttonLabels.getString("cancel"));
				cancelButton.setActionCommand("DIALOG_CANCEL");
				cancelButton.addActionListener(infoDialog);
				gbc.gridwidth = GridBagConstraints.REMAINDER;
				gbl.setConstraints(cancelButton, gbc);
				infoDialog.add(cancelButton);
				
			default:
				Button okButton = new Button(buttonLabels.getString("ok"));
				okButton.setActionCommand("DIALOG_YES");
				okButton.addActionListener(infoDialog);
				gbl.setConstraints(okButton, gbc);
				infoDialog.add(okButton);
				break;
		}
		
		infoDialog.addWindowListener(infoDialog);
		
		infoDialog.pack();
		JCards.centreWindow(parentFrame, infoDialog);
		infoDialog.show();
		infoDialog.dispose();
		
		return infoDialog.getResponse();
	}	
	
	/** User clicked a button */
	public void actionPerformed(ActionEvent e)
	{
		String ac = e.getActionCommand();
		
		if (ac.equals("DIALOG_YES"))
		{
			dispose();
			response = YES;
		}
		else if (ac.equals("DIALOG_NO"))
		{
			dispose();
			response = NO;
		}
		else if (ac.equals("DIALOG_CANCEL"))
		{
			dispose();
			response = CANCEL;
		}
	}
	
	/** Get the user's response */
	public int getResponse() {
		return response;
	}
	
	public void windowClosing(WindowEvent e)
	{
		response = CANCEL;
		dispose();
	}

	public void windowActivated(WindowEvent e) {}
	public void windowClosed(WindowEvent e) {}
	public void windowDeactivated(WindowEvent e) {}
	public void windowDeiconified(WindowEvent e) {}
	public void windowIconified(WindowEvent e) {}
	public void windowOpened(WindowEvent e) {}
	
}
