package nrs.jcards.io;

import java.io.*;
import java.text.MessageFormat;
import java.text.ParseException;
import java.util.Hashtable;
import java.util.StringTokenizer;

import nrs.jcards.*;
import nrs.jcards.dialog.InfoDialog;
import nrs.util.CSVListener;
import nrs.util.CSVParser;

/*
 * JCards
 * Copyright (C) 2004 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * CSV input/output
 *
 * @author Copyright &#169; 2002-2004 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class CSVIO implements CSVListener
{
	private JCards jcards = null;
	private Cards cards = null;
	private DBDef def = DBDef.getInstance();
	private CSVParser csvParser = null;
	private Hashtable csvFields = null;
	private int fieldCount = 0;
	private int numOfFields = 0;
	private int oldLineNum = -1;
	
	/** Creates a new instance of CSVIO */
	public CSVIO(JCards jcards)
	{
		this.jcards = jcards;
		cards = jcards.getCards();
	}

	/**
	 * Define the database's fields on the fly
	 * @param line line of fields
	 * @param fieldDelimiter character separating fields
	 */
	private void defineFields(String line, String fieldDelimiter)
	{
		String fieldName;
		FieldDef newFD;

		StringTokenizer st = new StringTokenizer(line,  fieldDelimiter);
		while (st.hasMoreTokens())
		{
			fieldName = st.nextToken();
			JCards.log("Defining new field "+fieldName, 5);
			newFD = new FieldDef(FieldDef.STRING, fieldName);
			def.addFieldDef(newFD);
		}

		numOfFields = def.getNumberOfFields();
		JCards.log("Defined "+numOfFields+" fields", 5);
	}
	
	/**
	 * Import a CSV file
	 * @param file the file to import
	 * @param fDelim field delimiter
	 * @param tDelim text delimiter
	 * @param charSet character set
	 * @param header if true parse the first line as containing field names
	 */
	public void importCSV(File file, String fDelim, String tDelim, String charSet, boolean header)
	{
		BufferedReader reader = null;
	
		try
		{
			reader = new BufferedReader(
				new InputStreamReader(new FileInputStream(file), charSet));
			// read header row
			if (header)
			{
				JCards.log("Reading header row", 5);
				String line = reader.readLine();
				if (line == null)
				{
					InfoDialog.showInfoDialog(jcards, jcards.getMessage("error"), InfoDialog.DIALOG_INFO, jcards.getMessage("eof"));
					return;
				}

				// automatically define fields
				if (def.getNumberOfFields() == 0)
				{
					JCards.log("Automatically defining fields", 5);
					defineFields(line, fDelim);
				}
			}
			else if (def.getNumberOfFields() == 0)
			{
				InfoDialog.showInfoDialog(jcards, jcards.getMessage("no.fields.def"), InfoDialog.DIALOG_INFO, jcards.getMessage("no.fields.def.msg"));
				return;
			}

			// read data
			JCards.log("Reading data", 5);
			csvParser = new CSVParser(reader, fDelim, tDelim, this);
			csvParser.parse();

			if (!csvParser.hasReadAll()) {
				InfoDialog.showInfoDialog(jcards, jcards.getMessage("error"), InfoDialog.DIALOG_INFO, jcards.getMessage("csv.read.failed"));
			}
		}
		catch (Exception ex)
		{
			String[] msglines = { jcards.getMessage("invalid.file"), ex.getMessage() };
			InfoDialog.showInfoDialog(jcards, jcards.getMessage("error"), InfoDialog.DIALOG_INFO, msglines);
			return;
		}
		finally
		{
			try	{
				if (reader != null) reader.close();
			} catch (Exception ex) {}
			reader = null;
		}
	}
	
	/**
	 * Export data to a CSV file
	 * @param file the file to export to
	 * @param fDelim field delimiter
	 * @param tDelim text delimiter
	 * @param charSet character set
	 * @param concatNL concatenate new lines
	 */
	public void exportCSV(File file, String fDelim, String tDelim, String charSet, boolean concatNL, boolean outputFieldNames)
	{
		try
		{
			PrintWriter writer = new PrintWriter(
				new BufferedWriter(
					new OutputStreamWriter(new FileOutputStream(file), charSet)));
			if (outputFieldNames) {
				writer.println(DBDef.getInstance().getCSV(fDelim));
				JCards.log("Written db definition", 5);
			}
			writer.println(jcards.getCards().getCSV(fDelim, tDelim, concatNL));
			JCards.log("Written data", 5);
			writer.close();
		} catch (IOException ioex) {
			InfoDialog.showInfoDialog(jcards, jcards.getMessage("error"), InfoDialog.DIALOG_INFO, ioex.getMessage());
			return;
		}
		MessageFormat msgFormat = new MessageFormat(jcards.getMessage("exported.to"));
		String[] args = { file.getName() };
		InfoDialog.showInfoDialog(jcards, jcards.getMessage("exported"), InfoDialog.DIALOG_INFO, msgFormat.format(args));
	}

	/**
	 * A new CSV value has been found
	 * @param field the new field
	 * @param lineNum the line number the field was initially read from
	 */
	public void newCSVValue(String value, int lineNum)
	{
		if (csvFields == null) {
			csvFields = new Hashtable(def.getNumberOfFields());
		}

		// check to see if there are too many fields
		if (oldLineNum == lineNum && fieldCount == 0) {
			MessageFormat msgFormat = new MessageFormat(jcards.getMessage("csv.too.many.fields"));
			Integer[] args = { new Integer(lineNum+1) };
			InfoDialog.showInfoDialog(jcards, jcards.getMessage("csv.read.failed"), InfoDialog.DIALOG_INFO, msgFormat.format(args));
			csvParser.stop();
			return;
		}
		
		FieldDef fd = def.getFieldDef(fieldCount++);
		try {
			Object val = FieldDef.parseValue(fd.getType(), value);
			if (val != null) {
				csvFields.put(fd.getName(), val);
			}
		} catch (ParseException pex) {
			jcards.log("Error parsing CSV file: "+pex.getMessage(), 0);
			csvParser.stop();
			csvParser = null;
			String[] msgLines = { jcards.getMessage("csv.parse.error"), pex.getMessage() };
			InfoDialog.showInfoDialog(jcards, jcards.getMessage("error"), InfoDialog.DIALOG_INFO, msgLines);
		}

		if (fieldCount == def.getNumberOfFields())
		{
			Card card = new Card(csvFields);
			cards.addCardSilently(card);
			csvFields = new Hashtable(def.getNumberOfFields());
			fieldCount = 0;
		}
		
		oldLineNum = lineNum;
	}

	/** No more values will be read from the CSV file */
	public void noMoreCSVValues()
	{
		cards.finishedLoad();
	}
}
