package nrs.jcards.io;

import java.io.*;
import java.util.Enumeration;
import java.util.zip.*;

import nanoxml.*;

import nrs.jcards.*;
import nrs.jcards.dialog.InfoDialog;
import nrs.jcards.dialog.PasswordDialog;
import nrs.encryption.Encryptor;

/*
 * JCards
 * Copyright (C) 2004 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * Class to read and write the JCards file format.
 * <p />
 * Encrypted files are:
 * <ol>
 * <li>zipped first to ensure a minimum file size</li>
 * <li>encrypted</li>
 * <li>written to the zip file</li>
 * </ol>
 *
 * Unencrypted files are written directly to the zip file as characters..
 *
 * @author Copyright Copyright &#169; 2002-2004 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class JCAReader
{
	private JCards jcards;

	/** Zipfile entry constants */
	private final static String CARD_DATA = "Card Data";
	private final static String DIGEST = "Digest";
	private final static String HEADER = "Header";

	/** Header constants */
	private final static String ENCRYPTED = "Encrypted: ";
	private final static String VERSION = "Version: ";
	private final static String CREATED = "Created: ";
	private final static String MODIFIED = "Modified: ";

	/** Encodings */
	private final static String ASCII = "ASCII";
	private final static String UTF8 = "UTF8";

	/** XML Elements */
	private final static String CARD_START = "<Card>";
	private final static String CARD_END = "</Card>";
	private final static String CARDS_START = "<Cards>";
	private final static String CARDS_END = "</Cards>";
	private final static String COLWIDTHS_START = "<ColWidths>";
	private final static String COLWIDTHS_END = "</ColWidths>";
	private final static String DBDEF_START = "<DBDef>";
	private final static String DBDEF_END = "</DBDef>";
	private final static String INFO_START = "<info>";
	private final static String INFO_END = "</info>";
	private final static String JCARDS_START = "<JCards>";
	private final static String JCARDS_END = "</JCards>";

	/** Construct the reader with a reference to the main class for logging, etc */
	public JCAReader(JCards jcards)
	{
		this.jcards = jcards;
	}

	/**
	 * Read the bytes from a zip file entry
	 * @param zf ZipFIle
	 * @param zeName name of the zip entry to read
	 * @exception IOException can't read from file
	 */
	private byte[] readZipBytes(ZipFile zf, String zeName) throws IOException
	{
		ZipEntry ze = zf.getEntry(zeName);
		if (ze == null)
		{
			InfoDialog.showInfoDialog(jcards, jcards.getMessage("error"), InfoDialog.DIALOG_INFO, jcards.getMessage("invalid.file"));
			return null;
		}

		return readInputStream(zf.getInputStream(ze));
	}

	/**
	 * Read bytes from an InputStream
	 * @param is InputStream
	 * @return bytes read from stream
	 * @exception IOException couldn't read from stream
	 */
	private byte[] readInputStream(InputStream is) throws IOException
	{
		BufferedInputStream bis = new BufferedInputStream(is);
		ByteArrayOutputStream baos = new ByteArrayOutputStream(bis.available());
		int b;
		int bytesRead;
		byte[] bytes = new byte[4096];
		while ((bytesRead = bis.read(bytes)) > -1)
		{
			baos.write(bytes, 0, bytesRead);
		}
		baos.flush();
		baos.close();
		bis.close();

		return baos.toByteArray();
	}

	/**
	 * Write bytes to a zip file
	 * @param zos ZipOutputStream
	 * @param zeName name of the zip entry
	 * @param bytes data to write
	 * @exception IOException can't write to file
	 */
	private void writeZipBytes(ZipOutputStream zos, String zeName, byte[] bytes) throws IOException
	{
		ZipEntry ze = new ZipEntry(zeName);
		zos.putNextEntry(ze);
		writeBytes(zos, bytes);
		zos.closeEntry();
	}

	private void writeBytes(OutputStream os, byte[] bytes) throws IOException
	{
		os.write(bytes, 0, bytes.length);
	}

	/**
	 * Save the data to the specified file
	 * @param saveFile file to save to
	 * @param password password to encrypt with
	 * @exception Exception save didn't work
	 */
	public void saveFile(File saveFile, String password) throws Exception
	{
		ZipOutputStream zos = new ZipOutputStream(new FileOutputStream(saveFile));
		ZipOutputStream encZos = null;
		BufferedWriter writer;
		Encryptor enc = null;
		ZipEntry ze;
		byte[] digest = null;

		// write header
		StringBuffer sb = new StringBuffer(VERSION);
		sb.append(JCards.VERSION);
		sb.append("\n");
		sb.append(ENCRYPTED);
		if (password.length() > 0) {
			sb.append("true");
		} else {
			sb.append("false");
		}
		sb.append("\n");
		sb.append(CREATED);
		sb.append(jcards.getCreationDate());
		sb.append("\n");
		sb.append(MODIFIED);
		sb.append(jcards.getModifiedDate());

		byte[] headerBytes = sb.toString().getBytes(ASCII);
		writeZipBytes(zos, HEADER, headerBytes);
		headerBytes = null;
		jcards.log("Written header", 7);

		// do we need to encrypt it
		if (password.length() > 0)
		{
			jcards.log("Encrypting", 3);
			// zip before encrypting to ensure a small file size
			enc = new Encryptor(password, zos, true);

			writeZipBytes(zos, DIGEST, enc.getDigest());
			jcards.log("Written digest", 5);

			ze = new ZipEntry(CARD_DATA);
			zos.putNextEntry(ze);

			encZos = new ZipOutputStream(enc);
			ZipEntry encZe = new ZipEntry("encrypted");
			encZos.putNextEntry(encZe);

			writer = new BufferedWriter(new OutputStreamWriter(encZos, UTF8));
		}
		else
		{
			ze = new ZipEntry(CARD_DATA);
			zos.putNextEntry(ze);
			writer = new BufferedWriter(new OutputStreamWriter(zos, UTF8));
		}

		writer.write(JCARDS_START);

		// write DBDef
		DBDef def = DBDef.getInstance();
		writer.write(def.getXML().toString());
		jcards.log("Written DBDef", 7);

		// write Cards
		writer.write(CARDS_START);
		Cards cards = jcards.getCards();
		Card card;
		Enumeration enm = cards.elements();
		while (enm.hasMoreElements())
		{
			card = (Card)enm.nextElement();
			writer.write(card.getXML().toString());
		}
		writer.write(CARDS_END);
		jcards.log("Written Cards", 7);

		// write ColWidths
		XMLElement colWidths = jcards.getColWidths();
		writer.write(colWidths.toString());
		jcards.log("Written ColWidths", 7);

		writer.write(JCARDS_END);
		writer.close();
		zos.close();
		jcards.log("Written file", 3);
	}

	/**
	 * Open the specified file
	 * @param saveFile file to open
	 * @exception Exception couldn't open file
	 */
	public void openFile(File saveFile) throws Exception
	{
		BufferedReader reader;
		boolean encrypted = false;
		boolean oldFile = false;
		String line, fileVersion;
		String dbImport = null;
		Encryptor enc = null;
		byte[] zipBytes;
		int startIndex, endIndex;

		ZipFile zf = new ZipFile(saveFile);
		// read header
		ZipEntry headZE = zf.getEntry(HEADER);
		// header isn't mandatory but should be present
		if (headZE != null)
		{
			jcards.log("Reading header", 5);
			reader = new BufferedReader(new InputStreamReader(zf.getInputStream(headZE), ASCII));
			while ((line = reader.readLine()) != null)
			{
				jcards.log(line, 9);
				if (line.startsWith(VERSION))
				{
					fileVersion = line.substring(9);
					if (fileVersion.startsWith("3")) {
						oldFile = true;
					}
				}
				else if (line.startsWith(ENCRYPTED))
				{
					encrypted = (Boolean.valueOf(line.substring(11))).booleanValue();
					if (encrypted)
					{
						// get the password
						jcards.log("File is encrypted", 7);
						PasswordDialog pd = new PasswordDialog(jcards, "", false);
						pd.pack();
						jcards.centreWindow(jcards, pd);
						pd.show();
					}
					else
					{
						jcards.setPassword("");
					}
				}
				else if (line.startsWith(CREATED))
				{
					try {
						jcards.setCreationDate(Long.parseLong(line.substring(9)));
					} catch (Exception ex) {}
				}
				else if (line.startsWith(MODIFIED))
				{
					try {
						jcards.setModifiedDate(Long.parseLong(line.substring(10)));
					} catch (Exception ex) {}
				}
			}
			reader.close();
			reader = null;

			// check if password is required
			if (encrypted)
			{
				if (jcards.getPassword().length() == 0)
				{
					InfoDialog.showInfoDialog(jcards, jcards.getMessage("no.password"), InfoDialog.DIALOG_INFO, jcards.getMessage("missing.password"));
					return;
				}
				else if (!oldFile)
				{
					enc = new Encryptor(jcards.getPassword());
					String digest = new String(readZipBytes(zf, DIGEST), UTF8);
					String pdigest = new String(enc.getDigest(), UTF8);
					if (!pdigest.equals(digest))
					{
						InfoDialog.showInfoDialog(jcards, jcards.getMessage("error"), InfoDialog.DIALOG_INFO, jcards.getMessage("incorrect.password"));
						return;
					}
				}
			}
			headZE = null;
		}
		jcards.log("Read header", 5);

		jcards.log("Reading db definition", 5);
		// read db definition
		zipBytes = readZipBytes(zf, CARD_DATA);
		zf.close();
		jcards.log("Read zip bytes", 7);
		if (encrypted)
		{
			jcards.log("Decrypting data", 7);

			// decrypt
			zipBytes = enc.decrypt(zipBytes);
			enc = null;

			if (oldFile) {
				// read using platform character set
				dbImport = new String(zipBytes);
			}
			else
			{
				// unzip
				ByteArrayInputStream bais = new ByteArrayInputStream(zipBytes);
				ZipInputStream zis = new ZipInputStream(bais);
				zis.getNextEntry();
				byte[] decrypted = readInputStream(zis);
				zis.close();
				// read using UTF8
				dbImport = new String(decrypted, UTF8);
			}

			jcards.log("Decrypted data", 5);
			jcards.log(dbImport, 9);
		}
		else
		{
			if (oldFile) {
				// read using platform character set
				dbImport = new String(zipBytes);
			} else {
				// read using UTF8
				dbImport = new String(zipBytes, UTF8);
			}
		}
		// clear as much memory as possible
		zipBytes = null;
		System.gc();
		jcards.log("Imported bytes", 5);

		// due to memory issues in parsing large files split out the
		// individual bits of the file and process one at a time
		XMLElement dataEl = new XMLElement();
		// DBDef
		startIndex = dbImport.indexOf(DBDEF_START);
		endIndex = dbImport.indexOf(DBDEF_END) + DBDEF_END.length();
		if (startIndex == -1 || endIndex == -1) {
			String[] msgLines = { jcards.getMessage("error.open"), jcards.getMessage("no.dbdef") };
			InfoDialog.showInfoDialog(jcards, jcards.getMessage("error"), InfoDialog.DIALOG_INFO, msgLines);
			return;
		}
		String dbDef = dbImport.substring(startIndex, endIndex);
		dbImport = dbImport.substring(endIndex);
		dataEl.parseString(dbDef);
		DBDef.getInstance().setXML(dataEl);
		jcards.log("Read DBDef", 7);

		// Cards
		startIndex = dbImport.indexOf(CARDS_START);
		endIndex = dbImport.indexOf(CARDS_END) + CARDS_END.length();
		if (startIndex == -1 || endIndex == -1) {
			String[] msgLines = { jcards.getMessage("error.open"), jcards.getMessage("no.cards") };
			InfoDialog.showInfoDialog(jcards, jcards.getMessage("error"), InfoDialog.DIALOG_INFO, msgLines);
		}
		String cardsXML = dbImport.substring(startIndex, endIndex);
		dbImport = dbImport.substring(endIndex);
		parseCards(cardsXML);
		jcards.log("Read Cards", 7);

		// ColWidths
		startIndex = dbImport.indexOf(COLWIDTHS_START);
		if (startIndex > -1)
		{
			endIndex = dbImport.indexOf(COLWIDTHS_END) + COLWIDTHS_END.length();
			String colWidths = dbImport.substring(startIndex, endIndex);
			dataEl = new XMLElement();
			dataEl.parseString(colWidths);
			jcards.setColWidths(dataEl);
		}
		jcards.log("Read ColWidths", 7);
		jcards.log("Parsed XML", 5);
	}

	/**
	 * Parse cards one at a time
	 * @param cardsXML unparsed XML String representing Cards
	 */
	private void parseCards(String cardsXML) throws Exception
	{
		String card;
		int startIndex, endIndex;
		XMLElement cardEl;
		int cardEndLength = CARD_END.length();
		Cards cards = jcards.getCards();

		// read info
		startIndex = cardsXML.indexOf(INFO_START);
		endIndex = cardsXML.indexOf(INFO_END);
		if (startIndex > -1)
		{
			String info = cardsXML.substring(startIndex+INFO_START.length(), endIndex+INFO_END.length());
			cardsXML.substring(endIndex+7);
			cards.setInfo(info);
		}

		// read Cards
		while ((startIndex = cardsXML.indexOf(CARD_START)) != -1)
		{
			endIndex = cardsXML.indexOf(CARD_END) + cardEndLength;
			card = cardsXML.substring(startIndex, endIndex);
			cardsXML = cardsXML.substring(endIndex);
			cardEl = new XMLElement();
			cardEl.parseString(card);
			cards.addCardSilently(cardEl);
		}

		cards.finishedLoad();
	}
}
