package nrs.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;

/*
 * NRSUtils
 * Copyright (C) 2004 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * <p>Parses a CSV file and informs a listener of new values.</p>
 * <p>To use, declare the class that wants to parse CSV values as implementing
 * the <em>nrs.util.CSVListener</em> interface along with implementations of the
 * <em>newCSVValue(String value, boolean newLine)</em> and <em>noMoreCSVValues()</em> methods. Then
 * create a new instance of CSVParser and pass in a BufferedReader which wraps the
 * CSV source, field and text delimiters and the listening class. Calling the
 * <em>parse()</em> method will start parsing. If you want to stop halfway through
 * the file then call the <em>stop()</em> method.</p>
 * <p>This class can also be called directly as <em>java nrs.util.CSVParser -csv file.csv</em>
 * to see how it works.</p>
 *
 * @author Copyright &#169; 2004 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class CSVParser implements CSVListener
{
	private CSVListener listener = null;
	private BufferedReader reader = null;
	private String fieldDelimiter = new String(",").intern();
	private String textDelimiter = new String("\"").intern();
	private boolean stop = false;
	private boolean readAll = false;
	private boolean textField = false;
	private int lineNum;
	private StringBuffer sb = new StringBuffer();
	private boolean noTextDelimiter = false;
	private boolean pauseOnValue = false;
	private final static String SPACE = " ";
	
	/**
	 * Create a new instance of CSVParser using default values for the field and text delimiter.
	 * The user will need to set the reader and listener before calling the parse method.
	 * @see #setReader(BufferedReader)
	 * @see #setListener(CSVListener)
	 */
	public CSVParser() {}

	/**
	 * Creates a new instance of CSVReader
	 * @param reader BufferedReader of the CSV source
	 * @param fieldDelimiter the field delimiter, e.g. ,
	 * @param textDelimiter the text delimiter, e.g. "
	 * @param listener the class that wants to be informed of CSV values
	 */
	public CSVParser(BufferedReader reader, String fieldDelimiter, String textDelimiter, CSVListener listener)
	{
		this.reader = reader;
		this.listener = listener;
		this.fieldDelimiter = fieldDelimiter;
		this.textDelimiter = textDelimiter;
	}

	/**
	 * Set the BufferedReader which will read the CSV source
	 * @param reader the CSV source
	 */
	public void setReader(BufferedReader reader) {
		this.reader = reader;
	}

	/**
	 * Get the current BufferedReader
	 * @return BufferedReader
	 */
	public BufferedReader getReader() {
		return reader;
	}

	/**
	 * Set the listener of CSV events
	 * @param listener CSVListener
	 */
	public void setListener(CSVListener listener) {
		this.listener = listener;
	}

	/**
	 * Get the currect CSVListener
	 * @return CSVListener
	 */
	public CSVListener getListener() {
		return listener;
	}

	/**
	 * Set the field delimiter
	 * @param fDelim the field delimiter character
	 */
	public void setFieldDelimiter(String fDelim) {
		this.fieldDelimiter = fDelim;
	}

	/**
	 * Get the currect field delimiter
	 * @return field delimiter
	 */
	public String getFieldDelimiter() {
		return fieldDelimiter;
	}

	/**
	 * Set the text delimiter
	 * @param tDelim the text delimiter character
	 */
	public void setTextDelimiter(String tDelim) {
		this.textDelimiter = tDelim;
	}

	/**
	 * Get the current text delimiter
	 * @return text delimiter
	 */
	public String getTextDelimiter() {
		return textDelimiter;
	}

	public boolean isNoTextDelimiter()
	{
		return noTextDelimiter;
	}

	/**
	 * If set the parser won't look for text delimiters.
	 * In this case the end of the field delimiter marks the start of the value.
	 * @param noTextDelimiter set to true to not scan for text delimiter
	 */
	public void setNoTextDelimiter(boolean noTextDelimiter)
	{
		this.noTextDelimiter = noTextDelimiter;
	}
	
	/**
	 * Parse the data from the specified BufferedReader
	 * @exception IOException an error occured reading the CSV source
	 */
	public void parse() throws IOException
	{
		if (reader == null || listener == null) {
			throw new IOException("Either the BufferedReader or CSVListener have not been set");
		}

		String line;
		lineNum = 0;
		while ((line = reader.readLine()) != null && !stop)
		{
			parseLine(line, lineNum++);
		}
		if (sb.length() > 0) {
			listener.newCSVValue(sb.toString(), lineNum++);
		}
		listener.noMoreCSVValues();
		readAll = true;
	}

	/**
	 * Parse a line at a time. The CSVUpdate object will be informed of new fields.
	 * @param line the line to parse
	 * @param lineNum line number
	 */
	private void parseLine(String line, int lineNum)
	{
		String s;

		if (textField && line.length() == 0) {
			sb.append("\n");
			return;
		}
		
		for (int i=0; i<line.length(); i++)
		{
			s = line.substring(i, i+1).intern();

			if (s.equals(textDelimiter))
			{
				textField = !textField;
				if (textField && line.length() == 1) {
					sb.append("\n");
				}
				continue;
			}

			if (s.equals(fieldDelimiter) && !textField)
			{
				listener.newCSVValue(sb.toString(), lineNum);
				sb = new StringBuffer();
			}
			else if (s.equals(SPACE))
			{
				if (textField || noTextDelimiter) {
					sb.append(s);
				}
			}
			else
			{
				sb.append(s);
			}

			if (stop) {
				break;
			}
		}

		if (sb.length() > 0) {
			if (textField) {
				sb.append("\n");
			} else {
				listener.newCSVValue(sb.toString(), lineNum);
				sb = new StringBuffer();
			}
		}
	}

	/** Tell the CSVReader not to continue parsing */
	public void stop() {
		stop = true;
	}

	/**
	 * Check to see if the CSV source has a) finished being parsed, and b) all values were read
	 * @return true if all values were parsed, false otherwise
	 */
	public boolean hasReadAll() {
		return readAll;
	}

	/** Pause after each value is found */
	private void setPauseOnValue(boolean pauseOnValue)
	{
		this.pauseOnValue = pauseOnValue;
	}
	
	/** A new CSV value has been found */
	public void newCSVValue(String value, int lineNum)
	{
		if (pauseOnValue)
		{
			System.out.print("CSV value on line "+lineNum+": '"+value+"'");
			BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));
			try {
				String s = reader.readLine();
				if (s.equalsIgnoreCase("q")) {
					System.exit(0);
				}
			} catch (Exception ex) {
				System.out.println(ex.getMessage());
				System.exit(-1);
			}
		}
		else
		{
			System.out.println("CSV value on line "+lineNum+": '"+value+"'");
		}
	}

	public void noMoreCSVValues() {}

	private static void printHelp()
	{
		System.out.println("CSVParser <arguments>");
		System.out.println(" -h   print this help");
		System.out.println(" -csv CSV file to parse (required)");
		System.out.println(" -f   field delimiter (default ,)");
		System.out.println(" -p   pause on each found value");
		System.out.println(" -t   text delimiter (default \")");
		System.out.println(" -nt no text delimiter");
		System.out.println("Use t for tab, s for space.\n");
		System.exit(0);
	}

	/**
	 * Main method for testing
	 * @param args command line arguments
	 */
	public static void main(String[] args)
	{
		String filename = null;
		String fDelim=",", tDelim="\"";
		String arg;
		boolean pause = false;
		boolean noTextDelimiter = false;

		try {
			for (int i=0; i<args.length; i++)
			{
				arg = args[i];

				if (arg.equals("-h") || arg.equals("--help")) {
					printHelp();
				} else if (arg.equals("-csv")) {
					filename = args[++i];
				}
				else if (arg.equals("-f"))
				{
					fDelim = args[++i];
					if (fDelim.length() > 1) {
						System.err.println("Field delimiter must be one character");
						System.exit(-1);
					}
					if (fDelim.equals("t")) {
						fDelim = "\t";
					}
				}
				else if (arg.equals("-nt"))
				{
					noTextDelimiter = true;
				}
				else if (arg.equals("-p"))
				{
					pause = true;
				}
				else if (arg.equals("-t"))
				{
					tDelim = args[++i];
					if (tDelim.length() > 1) {
						System.err.println("Text delimiter must be one character");
						System.exit(-1);
					}
					if (tDelim.equals("s")) {
						tDelim = " ";
					}
				}
			}
		} catch (IndexOutOfBoundsException iex) {
			System.err.println("Not enough arguments specified");
			printHelp();
		}

		if (filename == null) {
			System.err.println("Missing CSV filename");
			System.exit(-1);
		}

		try {
			File f = new File(filename);
			if (!f.exists()) {
				System.err.println("Cannot find "+filename);
				System.exit(-1);
			}
			BufferedReader reader = new BufferedReader(new FileReader(f));
			CSVParser parser = new CSVParser();
			parser.setReader(reader);
			parser.setListener(parser);
			parser.setFieldDelimiter(fDelim);
			parser.setTextDelimiter(tDelim);
			parser.setNoTextDelimiter(noTextDelimiter);
			if (pause)
			{
				parser.setPauseOnValue(pause);
				System.out.println("Press enter after to see each value. Type q to exit.");
			}
			parser.parse();
		} catch (Exception ex) {
			ex.printStackTrace();
		}
	}
}
